from __future__ import annotations

import logging
import subprocess
import typing as t
from typing import TYPE_CHECKING

from .base import Arguments
from .docker import ENV
from .docker import find_binary
from .docker import health as _docker_health

if TYPE_CHECKING:
    from ..types import PathType
    from .base import ArgType

logger = logging.getLogger(__name__)

__all__ = ["ENV", "health", "construct_build_args", "BUILDKIT_SUPPORT", "BUILD_CMD"]

BUILDKIT_SUPPORT = True

BUILD_CMD = ["buildx", "build"]


def health() -> bool:
    if not _docker_health():
        return False
    client = find_binary()
    assert client is not None
    has_buildx = subprocess.check_output([client, "buildx", "--help"]).decode("utf-8")
    if "--builder string" not in has_buildx:
        logger.warning(
            "Buildx is not installed. See https://docs.docker.com/build/buildx/install/ for instalation instruction."
        )
        return False
    return True


def parse_dict_opt(d: dict[str, str]) -> str:
    return ",".join([f"{key}={value}" for key, value in d.items()])


def construct_build_args(
    *,
    context_path: PathType = ".",
    add_host: dict[str, str] | ArgType = None,
    build_arg: dict[str, str] | ArgType = None,
    build_context: dict[str, str] | ArgType = None,
    cache_from: str | dict[str, str] | ArgType = None,
    cache_to: str | dict[str, str] | ArgType = None,
    label: dict[str, str] | ArgType = None,
    load: bool = True,
    no_cache_filter: str | dict[str, str] | ArgType = None,
    output: str | dict[str, str] | ArgType = None,
    platform: str | ArgType = None,
    push: bool = False,
    secret: str | dict[str, str] | ArgType = None,
    ulimit: str | dict[str, tuple[int, int]] | ArgType = None,
    **kwargs: t.Any,
) -> Arguments:
    cmds = Arguments()

    if platform and len(platform) > 1:
        if not push:
            logger.warning(
                "Multiple '--platform' arguments were found. Make sure to also use '--push' to push images to a repository or generated images will not be saved. See https://docs.docker.com/engine/reference/commandline/buildx_build/#load."
            )
    if push:
        load = False
    if isinstance(output, dict):
        output = parse_dict_opt(output)
    if output is not None and any("local" in o for o in output):
        load, push = False, False
    cmds.construct_args(output, opt="output")
    cmds.construct_args(push, opt="push")
    cmds.construct_args(load, opt="load")
    cmds.construct_args(platform, opt="platform")

    if isinstance(add_host, dict):
        add_host = tuple(f"{host}:{ip}" for host, ip in add_host.items())
    cmds.construct_args(add_host, opt="add-host")
    if isinstance(build_arg, dict):
        build_arg = tuple(f"{key}={value}" for key, value in build_arg.items())
    cmds.construct_args(build_arg, opt="build-arg")
    if isinstance(build_context, dict):
        build_context = tuple(f"{key}={value}" for key, value in build_context.items())
    cmds.construct_args(build_context, opt="build-context")
    if isinstance(cache_from, dict):
        cache_from = parse_dict_opt(cache_from)
    cmds.construct_args(cache_from, opt="cache-from")
    if isinstance(cache_to, dict):
        cache_to = parse_dict_opt(cache_to)
    cmds.construct_args(cache_to, opt="cache-to")
    if isinstance(label, dict):
        label = tuple(f"{key}={value}" for key, value in label.items())
    cmds.construct_args(label, opt="label")
    if isinstance(no_cache_filter, dict):
        no_cache_filter = parse_dict_opt(no_cache_filter)
    cmds.construct_args(no_cache_filter, opt="no-cache-filter")
    if isinstance(secret, dict):
        secret = parse_dict_opt(secret)
    cmds.construct_args(secret, opt="secret")
    if isinstance(ulimit, dict):
        # {"cpu": (1023, 1024), "fsize": (8192, 8192)}
        ulimit = tuple(f"{key}={value[0]}:{value[1]}" for key, value in ulimit.items())
    cmds.construct_args(ulimit, opt="ulimit")

    for k, v in kwargs.items():
        cmds.construct_args(v, opt=k.replace("_", "-"))

    cmds.append(str(context_path))

    return cmds



from __future__ import annotations

import typing as t

import click

if t.TYPE_CHECKING:
    TupleStrAny = tuple[str, ...]
    from bentoml._internal.cloud.schemas import DeploymentListSchema
    from bentoml._internal.cloud.schemas import DeploymentSchema

    from .utils import SharedOptions
else:
    TupleStrAny = tuple


def add_deployment_command(cli: click.Group) -> None:
    import json

    from rich.table import Table

    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.utils import bentoml_cattr
    from bentoml._internal.utils import rich_console as console
    from bentoml_cli.utils import BentoMLCommandGroup

    client = BentoMLContainer.bentocloud_client.get()
    output_option = click.option(
        "-o",
        "--output",
        type=click.Choice(["json", "default"]),
        default="default",
        help="Display the output of this command.",
    )

    def shared_decorator(f: t.Callable[..., t.Any]):
        options = [
            click.argument(
                "deployment-name",
                type=click.STRING,
                required=True,
            ),
            click.option(
                "--cluster-name",
                type=click.STRING,
                default=None,
                help="Name of the cluster.",
            ),
            click.option(
                "--kube-namespace",
                type=click.STRING,
                default=None,
                help="Kubernetes namespace.",
            ),
            output_option,
        ]
        for opt in reversed(options):
            f = opt(f)
        return f

    @cli.group(name="deployment", cls=BentoMLCommandGroup)
    def deployment_cli():
        """Deployment Subcommands Groups"""

    @deployment_cli.command()
    @click.option(
        "-f",
        "--file",
        type=click.File(),
        help="JSON file path for the deployment configuration",
    )
    @output_option
    @click.pass_obj
    def create(  # type: ignore
        shared_options: SharedOptions,
        file: str,
        output: t.Literal["json", "default"],
    ) -> DeploymentSchema:
        """Create a deployment on BentoCloud.

        \b
        A deployment can be created using a json file with configurations.
        The json file has the exact format as the one on BentoCloud Deployment UI.
        """
        res = client.deployment.create_from_file(
            path_or_stream=file, context=shared_options.cloud_context
        )
        if output == "default":
            console.print(res)
        elif output == "json":
            click.echo(json.dumps(bentoml_cattr.unstructure(res), indent=2))
        return res

    @deployment_cli.command()
    @click.option(
        "-f",
        "--file",
        type=click.File(),
        help="JSON file path for the deployment configuration",
    )
    @click.option("-n", "--name", type=click.STRING, help="Deployment name")
    @click.option("--bento", type=click.STRING, help="Bento tag")
    @output_option
    @click.pass_obj
    def update(  # type: ignore
        shared_options: SharedOptions,
        file: str | None,
        name: str | None,
        bento: str | None,
        output: t.Literal["json", "default"],
    ) -> DeploymentSchema:
        """Update a deployment on BentoCloud.

        \b
        A deployment can be updated using a json file with needed configurations.
        The json file has the exact format as the one on BentoCloud Deployment UI.
        """
        if file is not None:
            if name is not None:
                click.echo("Reading from file, ignoring --name", err=True)
            res = client.deployment.update_from_file(
                path_or_stream=file, context=shared_options.cloud_context
            )
        elif name is not None:
            res = client.deployment.update(
                name,
                bento=bento,
                context=shared_options.cloud_context,
                latest_bento=True,
            )
        else:
            raise click.BadArgumentUsage(
                "Either --file or --name is required for update command"
            )
        if output == "default":
            console.print(res)
        elif output == "json":
            unstructured = bentoml_cattr.unstructure(res)
            click.echo(json.dumps(unstructured, indent=2))
        return res

    @deployment_cli.command()
    @shared_decorator
    @click.pass_obj
    def get(  # type: ignore
        shared_options: SharedOptions,
        deployment_name: str,
        cluster_name: str,
        kube_namespace: str,
        output: t.Literal["json", "default"],
    ) -> DeploymentSchema:
        """Get a deployment on BentoCloud."""
        res = client.deployment.get(
            deployment_name=deployment_name,
            context=shared_options.cloud_context,
            cluster_name=cluster_name,
            kube_namespace=kube_namespace,
        )
        if output == "default":
            console.print(res)
        elif output == "json":
            unstructured = bentoml_cattr.unstructure(res)
            click.echo(json.dumps(unstructured, indent=2))
        return res

    @deployment_cli.command()
    @shared_decorator
    @click.pass_obj
    def terminate(  # type: ignore
        shared_options: SharedOptions,
        deployment_name: str,
        cluster_name: str,
        kube_namespace: str,
        output: t.Literal["json", "default"],
    ) -> DeploymentSchema:
        """Terminate a deployment on BentoCloud."""
        res = client.deployment.terminate(
            deployment_name=deployment_name,
            context=shared_options.cloud_context,
            cluster_name=cluster_name,
            kube_namespace=kube_namespace,
        )
        if output == "default":
            console.print(res)
        elif output == "json":
            unstructured = bentoml_cattr.unstructure(res)
            click.echo(json.dumps(unstructured, indent=2))
        return res

    @deployment_cli.command()
    @shared_decorator
    @click.pass_obj
    def delete(  # type: ignore
        shared_options: SharedOptions,
        deployment_name: str,
        cluster_name: str,
        kube_namespace: str,
        output: t.Literal["json", "default"],
    ) -> DeploymentSchema:
        """Delete a deployment on BentoCloud."""
        res = client.deployment.delete(
            deployment_name=deployment_name,
            context=shared_options.cloud_context,
            cluster_name=cluster_name,
            kube_namespace=kube_namespace,
        )
        if output == "default":
            console.print(res)
        elif output == "json":
            unstructured = bentoml_cattr.unstructure(res)
            click.echo(json.dumps(unstructured, indent=2))
        return res

    @deployment_cli.command()
    @click.option(
        "--cluster-name", type=click.STRING, default=None, help="Name of the cluster."
    )
    @click.option(
        "--query", type=click.STRING, default=None, help="Query for list request."
    )
    @click.option(
        "--search", type=click.STRING, default=None, help="Search for list request."
    )
    @click.option(
        "--start", type=click.STRING, default=None, help="Start for list request."
    )
    @click.option(
        "--count", type=click.STRING, default=None, help="Count for list request."
    )
    @click.option(
        "-o",
        "--output",
        help="Display the output of this command.",
        type=click.Choice(["json", "default", "table"]),
        default="table",
    )
    @click.pass_obj
    def list(  # type: ignore
        shared_options: SharedOptions,
        cluster_name: str,
        query: str,
        search: str,
        count: int,
        start: int,
        output: t.Literal["json", "default", "table"],
    ) -> DeploymentListSchema:
        """List existing deployments on BentoCloud."""
        res = client.deployment.list(
            context=shared_options.cloud_context,
            cluster_name=cluster_name,
            query=query,
            search=search,
            count=count,
            start=start,
        )
        if output == "table":
            table = Table(box=None)
            table.add_column("Deployment")
            table.add_column("Bento")
            table.add_column("Status")
            table.add_column("Created At")
            for deployment in res.items:
                target = deployment.latest_revision.targets[0]
                table.add_row(
                    deployment.name,
                    f"{target.bento.repository.name}:{target.bento.name}",
                    deployment.status.value,
                    deployment.created_at.astimezone().strftime("%Y-%m-%d %H:%M:%S"),
                )
            console.print(table)
        elif output == "default":
            console.print(res)
        elif output == "json":
            unstructured = bentoml_cattr.unstructure(res)
            click.echo(json.dumps(unstructured, indent=2))
        return res



"""
This module is shim for bentoctl. NOT FOR DIRECT USE.
"""
from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

import fs

from ..container.generate import generate_containerfile

if TYPE_CHECKING:
    from .build_config import DockerOptions

__all__ = ["generate_dockerfile"]

logger = logging.getLogger(__name__)

warnings.warn(
    "%s is deprecated. Make sure to use 'bentoml.container.build' and 'bentoml.container.health' instead."
    % __name__,
    DeprecationWarning,
    stacklevel=4,
)


def generate_dockerfile(docker: DockerOptions, context_path: str, *, use_conda: bool):
    from ..bento import Bento

    bento = Bento.from_fs(fs.open_fs(context_path))
    logger.debug("'use_conda' is deprecated and will not be used.")
    return generate_containerfile(
        docker,
        bento.path,
        conda=bento.info.conda,
        bento_fs=bento._fs,
    )



# -*- coding: utf-8 -*-
""" Unflatten nested dict/array data
*** This is a modified version of the original unflatten.py from https://github.com/dairiki/unflatten, which is
published under the license ***

Copyright (C) 2018 Geoffrey T. Dairiki
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

3. The name of the author may not be used to endorse or promote
   products derived from this software without specific prior written
   permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

"""
from __future__ import absolute_import
from __future__ import annotations

import re
import sys
import typing as t
from operator import itemgetter

if sys.version_info[0] == 2:
    string_type = basestring  # noqa: F821
else:
    string_type = str


def unflatten(arg: dict[str, t.Any]) -> dict[str, t.Any]:
    """Unflatten nested dict/array data.

    This function takes a single argument which may either be a
    ``dict`` (or any object having a dict-like ``.items()`` or
    ``.iteritems()`` method) or a sequence of ``(key, value)`` pairs.
    The keys in the ``dict`` or sequence should must all be strings.

    Examples
    --------

    Nested ``dict``\\s::

    >>> unflatten({'foo.bar': 'val'})
    {'foo': {'bar': 'val'}}

    Nested ``list``::

    >>> unflatten({'foo[0]': 'val', 'foo[1]': 'bar'})
    {'foo': ['val', 'bar']}

    Nested ``list``\\s::

    >>> unflatten({'foo[0][0]': 'val'})
    {'foo': [['val']]}

    Lists of ``dict``\\s::

    >>> unflatten({'foo[0].bar': 'val',
    ...            'foo[1].baz': 'x'})
    {'foo': [{'bar': 'val'}, {'baz': 'x'}]}

    """
    if hasattr(arg, "items"):
        items = arg.items()
    else:
        items = arg

    data: dict[str, t.Any] = {}
    holders: list[t.Any] = []
    for flat_key, val in items:
        parsed_key = _parse_key(flat_key)
        obj = data
        for depth, (key, next_key) in enumerate(zip(parsed_key, parsed_key[1:]), 1):
            if isinstance(next_key, string_type):
                holder_type = _dict_holder
            else:
                holder_type = _list_holder

            if key not in obj:
                obj[key] = holder_type(_unparse_key(parsed_key[:depth]))
                holders.append((obj, key))
            elif not isinstance(obj[key], holder_type):
                raise ValueError(
                    "conflicting types %s and %s for key %r"
                    % (
                        _node_type(obj[key]),
                        holder_type.node_type,
                        _unparse_key(parsed_key[:depth]),
                    )
                )
            obj = obj[key]

        last_key = parsed_key[-1]
        if isinstance(obj.get(last_key), _holder):
            raise ValueError(
                "conflicting types %s and terminal for key %r"
                % (_node_type(obj[last_key]), flat_key)
            )
        obj[last_key] = val

    for obj, key in reversed(holders):
        obj[key] = obj[key].getvalue()

    return data


def _node_type(value: _holder) -> tuple[object] | t.Literal["terminal"]:
    if isinstance(value, _holder):
        return (value.node_type,)
    else:
        return "terminal"


class _holder(dict):
    node_type: type

    def __init__(self, flat_key: str):
        self.flat_key = flat_key
        self.data: dict[t.Any, t.Any] = {}

    def __contains__(self, key: t.Any):
        return key in self.data

    def __getitem__(self, key: t.Any):
        return self.data[key]

    def get(self, key: t.Any):
        return self.data.get(key)

    def __setitem__(self, key: t.Any, value: t.Any):
        self.data[key] = value


class _dict_holder(_holder):
    node_type = dict

    def getvalue(self):
        return self.data


class _list_holder(_holder):
    node_type = list

    def getvalue(self) -> list[t.Any]:
        items = sorted(self.data.items(), key=itemgetter(0))
        value: list[t.Any] = []
        for n, (key, val) in enumerate(items):
            if key != n:
                assert key > n
                missing_key = "%s[%d]" % (self.flat_key, n)
                raise ValueError("missing key %r" % missing_key)
            value.append(val)
        return value


_dot_or_indexes_re = re.compile(r"(\.?\"[^\"]*\")|(\[\d+\])|(\.\w*)|(^\w*)")


def _parse_key(flat_key: str):
    if not isinstance(flat_key, string_type):
        raise TypeError("keys must be strings")

    split_key = _dot_or_indexes_re.split(flat_key)
    parts: list[t.Any] = [""] if flat_key.startswith(".") else []

    for i in range(0, len(split_key), 5):
        sep = split_key[i]

        if sep != "":
            raise ValueError("invalid separator %r in key %r" % (sep, flat_key))

        if len(split_key) < i + 4:
            break

        if split_key[i + 1] is not None:
            # quoted first string
            string = split_key[i + 1]
            if len(string) == 2:
                string = ""
            elif i == 0:
                string = string[2:-1] if string.startswith(".") else string[1:-1]
            else:
                if string[0] != ".":
                    raise ValueError("invalid string %r in key %r" % (string, flat_key))
                string = string[2:-1]
            parts.append(string)

        elif split_key[i + 2] is not None:
            # index
            parts.append(int(split_key[i + 2][1:-1]))

        elif split_key[i + 3] is not None or split_key[i + 4] is not None:
            # unquoted string
            string = split_key[i + 3] or split_key[i + 4] or ""
            if len(string) == 0:
                string = ""
            elif i == 0:
                string = string[1:] if string.startswith(".") else string
            else:
                if string[0] != ".":
                    raise ValueError("invalid string %r in key %r" % (string, flat_key))
                string = string[1:]
            parts.append(string)
        else:
            assert False

    if len(parts) > 0 and isinstance(parts[0], int):
        parts.insert(0, "")
    return parts


def _unparse_key(parsed: list[t.Any]) -> str:
    bits: list[str] = []
    for part in parsed:
        if isinstance(part, string_type):
            if part.isidentifier():
                fmt = ".%s" if bits else "%s"
            elif part == "":
                fmt = ".%s" if bits else "%s"
            else:
                fmt = '."%s"' if bits else '"%s"'
        else:
            fmt = "[%d]"
        bits.append(fmt % part)
    return "".join(bits)



from __future__ import annotations

import logging
import os
import sys
import typing as t

import click

if t.TYPE_CHECKING:
    P = t.ParamSpec("P")
    F = t.Callable[P, t.Any]


logger = logging.getLogger(__name__)

DEFAULT_DEV_SERVER_HOST = "127.0.0.1"


def deprecated_option(*param_decls: str, **attrs: t.Any):
    """Marks a given options as deprecated, and omit a warning when it's used"""
    deprecated = attrs.pop("deprecated", True)
    new_behaviour = attrs.pop("current_behaviour", None)
    assert new_behaviour is not None, "current_behaviour is required"

    def show_deprecated_callback(
        ctx: click.Context, param: click.Parameter, value: t.Any
    ):
        if value is not param.default and deprecated:
            name = "'--%(name)s'" if attrs.get("is_flag", False) else "'%(name)s'"
            DEPRECATION_WARNING = f"DeprecationWarning: The parameter {name} is deprecated and will be removed in the future. (Current behaviour: %(new_behaviour)s)"
            click.secho(
                DEPRECATION_WARNING
                % {"name": param.name, "new_behaviour": new_behaviour},
                fg="yellow",
                err=True,
            )

    def decorator(f: F[t.Any]) -> t.Callable[[F[t.Any]], click.Command]:
        msg = attrs.pop("help", "")
        msg += " (Deprecated)" if msg else "(Deprecated)"
        attrs.setdefault("help", msg)
        attrs.setdefault("callback", show_deprecated_callback)
        return click.option(*param_decls, **attrs)(f)

    return decorator


def add_serve_command(cli: click.Group) -> None:
    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.log import configure_server_logging
    from bentoml.grpc.utils import LATEST_PROTOCOL_VERSION
    from bentoml_cli.env_manager import env_manager

    @cli.command(aliases=["serve-http"])
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--development",
        type=click.BOOL,
        help="Run the BentoServer in development mode",
        is_flag=True,
        default=False,
        show_default=True,
    )
    @deprecated_option(
        "--production",
        type=click.BOOL,
        help="Run BentoServer in production mode",
        current_behaviour="This is enabled by default. To run in development mode, use '--development'.",
        is_flag=True,
        default=True,
        show_default=False,
    )
    @click.option(
        "-p",
        "--port",
        type=click.INT,
        default=BentoMLContainer.http.port.get(),
        help="The port to listen on for the REST api server",
        envvar="BENTOML_PORT",
        show_default=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        default=BentoMLContainer.http.host.get(),
        help="The host to bind for the REST api server",
        envvar="BENTOML_HOST",
        show_default=True,
    )
    @click.option(
        "--api-workers",
        type=click.INT,
        default=BentoMLContainer.api_server_workers.get(),
        help="Specify the number of API server workers to start. Default to number of available CPU cores in production mode",
        envvar="BENTOML_API_WORKERS",
        show_default=True,
    )
    @click.option(
        "--timeout",
        type=click.INT,
        help="Specify the timeout (seconds) for API server and runners",
        envvar="BENTOML_TIMEOUT",
    )
    @click.option(
        "--backlog",
        type=click.INT,
        default=BentoMLContainer.api_server_config.backlog.get(),
        help="The maximum number of pending connections.",
        show_default=True,
    )
    @click.option(
        "--reload",
        type=click.BOOL,
        is_flag=True,
        help="Reload Service when code changes detected",
        default=False,
        show_default=True,
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--ssl-certfile",
        type=str,
        default=BentoMLContainer.ssl.certfile.get(),
        help="SSL certificate file",
        show_default=True,
    )
    @click.option(
        "--ssl-keyfile",
        type=str,
        default=BentoMLContainer.ssl.keyfile.get(),
        help="SSL key file",
        show_default=True,
    )
    @click.option(
        "--ssl-keyfile-password",
        type=str,
        default=BentoMLContainer.ssl.keyfile_password.get(),
        help="SSL keyfile password",
        show_default=True,
    )
    @click.option(
        "--ssl-version",
        type=int,
        default=BentoMLContainer.ssl.version.get(),
        help="SSL version to use (see stdlib 'ssl' module)",
        show_default=True,
    )
    @click.option(
        "--ssl-cert-reqs",
        type=int,
        default=BentoMLContainer.ssl.cert_reqs.get(),
        help="Whether client certificate is required (see stdlib 'ssl' module)",
        show_default=True,
    )
    @click.option(
        "--ssl-ca-certs",
        type=str,
        default=BentoMLContainer.ssl.ca_certs.get(),
        help="CA certificates file",
        show_default=True,
    )
    @click.option(
        "--ssl-ciphers",
        type=str,
        default=BentoMLContainer.ssl.ciphers.get(),
        help="Ciphers to use (see stdlib 'ssl' module)",
        show_default=True,
    )
    @env_manager
    def serve(  # type: ignore (unused warning)
        bento: str,
        development: bool,
        port: int,
        host: str,
        api_workers: int | None,
        timeout: int | None,
        backlog: int,
        reload: bool,
        working_dir: str | None,
        ssl_certfile: str | None,
        ssl_keyfile: str | None,
        ssl_keyfile_password: str | None,
        ssl_version: int | None,
        ssl_cert_reqs: int | None,
        ssl_ca_certs: str | None,
        ssl_ciphers: str | None,
        **attrs: t.Any,
    ) -> None:
        """Start a HTTP BentoServer from a given 🍱

        \b
        BENTO is the serving target, it can be the import as:
        - the import path of a 'bentoml.Service' instance
        - a tag to a Bento in local Bento store
        - a folder containing a valid 'bentofile.yaml' build file with a 'service' field, which provides the import path of a 'bentoml.Service' instance
        - a path to a built Bento (for internal & debug use only)

        e.g.:

        \b
        Serve from a bentoml.Service instance source code (for development use only):
            'bentoml serve fraud_detector.py:svc'

        \b
        Serve from a Bento built in local store:
            'bentoml serve fraud_detector:4tht2icroji6zput3suqi5nl2'
            'bentoml serve fraud_detector:latest'

        \b
        Serve from a Bento directory:
            'bentoml serve ./fraud_detector_bento'

        \b
        If '--reload' is provided, BentoML will detect code and model store changes during development, and restarts the service automatically.

        \b
        The '--reload' flag will:
        - be default, all file changes under '--working-dir' (default to current directory) will trigger a restart
        - when specified, respect 'include' and 'exclude' under 'bentofile.yaml' as well as the '.bentoignore' file in '--working-dir', for code and file changes
        - all model store changes will also trigger a restart (new model saved or existing model removed)
        """
        configure_server_logging()
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."
        if sys.path[0] != working_dir:
            sys.path.insert(0, working_dir)

        from bentoml.serve import serve_http_production

        if development:
            serve_http_production(
                bento,
                working_dir=working_dir,
                port=port,
                host=DEFAULT_DEV_SERVER_HOST if not host else host,
                backlog=backlog,
                api_workers=1,
                timeout=timeout,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_keyfile_password=ssl_keyfile_password,
                ssl_version=ssl_version,
                ssl_cert_reqs=ssl_cert_reqs,
                ssl_ca_certs=ssl_ca_certs,
                ssl_ciphers=ssl_ciphers,
                reload=reload,
                development_mode=True,
            )
        else:
            serve_http_production(
                bento,
                working_dir=working_dir,
                port=port,
                host=host,
                api_workers=api_workers,
                timeout=timeout,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_keyfile_password=ssl_keyfile_password,
                ssl_version=ssl_version,
                ssl_cert_reqs=ssl_cert_reqs,
                ssl_ca_certs=ssl_ca_certs,
                ssl_ciphers=ssl_ciphers,
                reload=reload,
                development_mode=False,
            )

    @cli.command(name="serve-grpc")
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--development",
        type=click.BOOL,
        help="Run the BentoServer in development mode",
        is_flag=True,
        default=False,
        show_default=True,
    )
    @deprecated_option(
        "--production",
        type=click.BOOL,
        help="Run BentoServer in production mode",
        current_behaviour="This is enabled by default. To run in development mode, use '--development'.",
        is_flag=True,
        default=True,
        show_default=False,
    )
    @click.option(
        "-p",
        "--port",
        type=click.INT,
        default=BentoMLContainer.grpc.port.get(),
        help="The port to listen on for the REST api server",
        envvar="BENTOML_PORT",
        show_default=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        default=BentoMLContainer.grpc.host.get(),
        help="The host to bind for the gRPC server",
        envvar="BENTOML_HOST",
        show_default=True,
    )
    @click.option(
        "--api-workers",
        type=click.INT,
        default=BentoMLContainer.api_server_workers.get(),
        help="Specify the number of API server workers to start. Default to number of available CPU cores in production mode",
        envvar="BENTOML_API_WORKERS",
        show_default=True,
    )
    @click.option(
        "--reload",
        type=click.BOOL,
        is_flag=True,
        help="Reload Service when code changes detected",
        default=False,
        show_default=True,
    )
    @click.option(
        "--backlog",
        type=click.INT,
        default=BentoMLContainer.api_server_config.backlog.get(),
        help="The maximum number of pending connections.",
        show_default=True,
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--enable-reflection",
        is_flag=True,
        default=BentoMLContainer.grpc.reflection.enabled.get(),
        type=click.BOOL,
        help="Enable reflection.",
        show_default=True,
    )
    @click.option(
        "--enable-channelz",
        is_flag=True,
        default=BentoMLContainer.grpc.channelz.enabled.get(),
        type=click.BOOL,
        help="Enable Channelz. See https://github.com/grpc/proposal/blob/master/A14-channelz.md.",
    )
    @click.option(
        "--max-concurrent-streams",
        default=BentoMLContainer.grpc.max_concurrent_streams.get(),
        type=click.INT,
        help="Maximum number of concurrent incoming streams to allow on a http2 connection.",
        show_default=True,
    )
    @click.option(
        "--ssl-certfile",
        type=str,
        default=BentoMLContainer.ssl.certfile.get(),
        help="SSL certificate file",
        show_default=True,
    )
    @click.option(
        "--ssl-keyfile",
        type=str,
        default=BentoMLContainer.ssl.keyfile.get(),
        help="SSL key file",
        show_default=True,
    )
    @click.option(
        "--ssl-ca-certs",
        type=str,
        default=BentoMLContainer.ssl.ca_certs.get(),
        help="CA certificates file",
        show_default=True,
    )
    @click.option(
        "-pv",
        "--protocol-version",
        type=click.Choice(["v1", "v1alpha1"]),
        help="Determine the version of generated gRPC stubs to use.",
        default=LATEST_PROTOCOL_VERSION,
        show_default=True,
    )
    @env_manager
    def serve_grpc(  # type: ignore (unused warning)
        bento: str,
        development: bool,
        port: int,
        host: str,
        api_workers: int | None,
        backlog: int,
        reload: bool,
        working_dir: str | None,
        ssl_certfile: str | None,
        ssl_keyfile: str | None,
        ssl_ca_certs: str | None,
        enable_reflection: bool,
        enable_channelz: bool,
        max_concurrent_streams: int | None,
        protocol_version: str,
        **attrs: t.Any,
    ):
        """Start a gRPC BentoServer from a given 🍱

        \b
        BENTO is the serving target, it can be the import as:
        - the import path of a 'bentoml.Service' instance
        - a tag to a Bento in local Bento store
        - a folder containing a valid 'bentofile.yaml' build file with a 'service' field, which provides the import path of a 'bentoml.Service' instance
        - a path to a built Bento (for internal & debug use only)

        e.g.:

        \b
        Serve from a bentoml.Service instance source code (for development use only):
            'bentoml serve-grpc fraud_detector.py:svc'

        \b
        Serve from a Bento built in local store:
            'bentoml serve-grpc fraud_detector:4tht2icroji6zput3suqi5nl2'
            'bentoml serve-grpc fraud_detector:latest'

        \b
        Serve from a Bento directory:
            'bentoml serve-grpc ./fraud_detector_bento'

        If '--reload' is provided, BentoML will detect code and model store changes during development, and restarts the service automatically.

        \b
        The '--reload' flag will:
        - be default, all file changes under '--working-dir' (default to current directory) will trigger a restart
        - when specified, respect 'include' and 'exclude' under 'bentofile.yaml' as well as the '.bentoignore' file in '--working-dir', for code and file changes
        - all model store changes will also trigger a restart (new model saved or existing model removed)
        """
        configure_server_logging()
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."

        from bentoml.serve import serve_grpc_production

        if development:
            serve_grpc_production(
                bento,
                working_dir=working_dir,
                port=port,
                host=DEFAULT_DEV_SERVER_HOST if not host else host,
                backlog=backlog,
                api_workers=1,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_ca_certs=ssl_ca_certs,
                max_concurrent_streams=max_concurrent_streams,
                reflection=enable_reflection,
                channelz=enable_channelz,
                protocol_version=protocol_version,
                reload=reload,
                development_mode=True,
            )
        else:
            serve_grpc_production(
                bento,
                working_dir=working_dir,
                port=port,
                backlog=backlog,
                host=host,
                api_workers=api_workers,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_ca_certs=ssl_ca_certs,
                max_concurrent_streams=max_concurrent_streams,
                reflection=enable_reflection,
                channelz=enable_channelz,
                protocol_version=protocol_version,
                reload=reload,
                development_mode=False,
            )



from __future__ import annotations

import functools
import typing as t

from ...exceptions import MissingDependencyException
from ..runner import Runner
from ..runner.runner import RunnerMethod
from ..runner.runner_handle import RunnerHandle

try:
    import ray
    from ray.serve.deployment import Deployment as RayDeployment
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'ray' is required in order to use module 'bentoml.ray', install with 'pip install -U \"ray[serve]\"'. For more information, refer to https://docs.ray.io/",
    )

if t.TYPE_CHECKING:
    P = t.ParamSpec("P")
else:
    P = t.TypeVar("P")

R = t.TypeVar("R")


class RayRunnerHandle(RunnerHandle):
    def __init__(self, runner: Runner, ray_deployment: RayDeployment) -> None:
        self._runner = runner
        self._ray_deployment = ray_deployment

    async def is_ready(self, timeout: int) -> bool:
        return True

    def run_method(
        self,
        __bentoml_method: RunnerMethod[t.Any, P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R | tuple[R, ...]:
        import anyio

        return t.cast(
            "R | tuple[R, ...]",
            anyio.from_thread.run(
                functools.partial(self.async_run_method, **kwargs),
                __bentoml_method,
                *args,
            ),
        )

    async def async_run_method(
        self,
        __bentoml_method: RunnerMethod[t.Any, P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R | tuple[R, ...]:
        return ray.get(
            await getattr(self._ray_deployment, __bentoml_method.name).remote(
                *args, **kwargs
            )
        )



from __future__ import annotations

import typing as t
from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from bentoml._internal import external_typing as ext


class GenerateLatestMiddleware:
    def __init__(self, app: ext.ASGIApp):
        from bentoml._internal.configuration.containers import BentoMLContainer

        self.app = app
        self.metrics_client = BentoMLContainer.metrics_client.get()

    async def __call__(
        self, scope: ext.ASGIScope, receive: ext.ASGIReceive, send: ext.ASGISend
    ) -> None:
        assert scope["type"] == "http"
        assert scope["path"] == "/"

        from starlette.responses import Response

        return await Response(
            self.metrics_client.generate_latest(),
            status_code=200,
            media_type=self.metrics_client.CONTENT_TYPE_LATEST,
        )(scope, receive, send)


@click.command()
@click.option("--fd", type=click.INT, required=True)
@click.option("--backlog", type=click.INT, default=2048)
@click.option(
    "--prometheus-dir",
    type=click.Path(exists=True),
    help="Required by prometheus to pass the metrics in multi-process mode",
)
def main(fd: int, backlog: int, prometheus_dir: str | None):
    """
    Start a standalone Prometheus server to use with gRPC.
    \b
    This is an internal API, users should not use this directly. Instead use 'bentoml serve-grpc'.
    Prometheus then can be accessed at localhost:9090
    """

    import socket

    import psutil
    import uvicorn
    from starlette.applications import Starlette
    from starlette.middleware import Middleware
    from starlette.middleware.wsgi import WSGIMiddleware  # TODO: a2wsgi

    from bentoml._internal.configuration import get_debug_mode
    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.context import component_context
    from bentoml._internal.log import configure_server_logging

    component_context.component_type = "prom_server"

    configure_server_logging()

    BentoMLContainer.development_mode.set(False)
    metrics_client = BentoMLContainer.metrics_client.get()
    if prometheus_dir is not None:
        BentoMLContainer.prometheus_multiproc_dir.set(prometheus_dir)

    # create a ASGI app that wraps around the default HTTP prometheus server.
    prom_app = Starlette(
        debug=get_debug_mode(), middleware=[Middleware(GenerateLatestMiddleware)]
    )
    prom_app.mount("/", WSGIMiddleware(metrics_client.make_wsgi_app()))
    sock = socket.socket(fileno=fd)

    uvicorn_options: dict[str, t.Any] = {
        "backlog": backlog,
        "log_config": None,
        "workers": 1,
    }
    if psutil.WINDOWS:
        uvicorn_options["loop"] = "asyncio"
        import asyncio

        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore
    uvicorn.Server(uvicorn.Config(prom_app, **uvicorn_options)).run(sockets=[sock])


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



from __future__ import annotations

import logging
import typing as t
from types import ModuleType

import cloudpickle

import bentoml

from ...exceptions import MissingDependencyException
from ...exceptions import NotFound
from ..models.model import Model
from ..models.model import ModelContext
from ..models.model import ModelOptions
from ..models.model import ModelSignature
from ..tag import Tag
from ..utils.pkg import get_pkg_version
from .common.pytorch import PyTorchTensorContainer  # noqa # type: ignore

try:
    import easyocr
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'easyocr' is required in order to use module 'bentoml.easyocr'. Install easyocr with 'pip install easyocr'."
    )

if t.TYPE_CHECKING:
    from ..models.model import ModelSignaturesType

    ListStr = list[str]
else:
    ListStr = list


__all__ = ["load_model", "save_model", "get_runnable", "get"]

MODULE_NAME = "bentoml.easyocr"
API_VERSION = "v1"
MODEL_FILENAME = "saved_model.pkl"


logger = logging.getLogger(__name__)


def get(tag_like: str | Tag) -> Model:
    """
    Get the BentoML model with the given tag.

    Args:
        tag_like: The tag of the model to retrieve from the model store.

    Returns:
        :obj:`~bentoml.Model`: A BentoML :obj:`~bentoml.Model` with the matching tag.

    Example:

    .. code-block:: python

       import bentoml
       # target model must be from the BentoML model store
       model = bentoml.easyocr.get("en_reader:latest")
    """
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def load_model(bento_model: str | Tag | Model) -> easyocr.Reader:
    """
    Load the EasyOCR model from BentoML local model store with given name.

    Args:
        bento_model: Either the tag of the model to get from the store,
                     or a BentoML :class:`~bentoml.Model` instance to load the
                     model from.

    Returns:
        ``easyocr.Reader``: The EasyOCR model from the model store.

    Example:

    .. code-block:: python

        import bentoml
        reader = bentoml.easyocr.load_model('en_reader:latest')
    """
    if not isinstance(bento_model, Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )
    with open(bento_model.path_of(MODEL_FILENAME), "rb") as f:
        return cloudpickle.load(f)


def save_model(
    name: Tag | str,
    reader: easyocr.Reader,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name: Name for given model instance. This should pass Python identifier check.
        reader: The EasyOCR model to be saved. Currently only supports pre-trained models from easyocr.
                Custom models are not yet supported.
        signatures: Methods to expose for running inference on the target model. Signatures are used for creating :obj:`~bentoml.Runner` instances when serving model with :obj:`~bentoml.Service`
        labels: User-defined labels for managing models, e.g. ``team=nlp``, ``stage=dev``.
        custom_objects: Custom objects to be saved with the model. An example is ``{"my-normalizer": normalizer}``.
                        Custom objects are currently serialized with cloudpickle, but this implementation is subject to change.
        external_modules: user-defined additional python modules to be saved alongside the model or custom objects,
                          e.g. a tokenizer module, preprocessor module, model configuration module
        metadata: Custom metadata for given model.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format ``name:version`` where ``name`` is the user-defined model's name, and a generated ``version``.

    Examples:

    .. code-block:: python

        import bentoml
        import easyocr

        reader = easyocr.Reader(['en'])
        bento_model = bentoml.easyocr.save_model('en_reader', reader)
    """  # noqa
    context = ModelContext(
        framework_name="easyocr",
        framework_versions={"easyocr": get_pkg_version("easyocr")},
    )

    if signatures is None:
        signatures = {
            k: {"batchable": False}
            for k in ("detect", "readtext", "readtextlang", "recognize")
        }
        signatures["readtext_batched"] = {"batchable": True}
        logger.info(
            'Using the default model signature for Transformers (%s) for model "%s".',
            signatures,
            name,
        )

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        labels=labels,
        context=context,
        options=ModelOptions(),
        signatures=signatures,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
    ) as bento_model:
        with open(bento_model.path_of(MODEL_FILENAME), "wb") as f:
            cloudpickle.dump(reader, f)
        return bento_model


def get_runnable(bento_model: bentoml.Model) -> type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class EasyOCRRunnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()

            self.model = load_model(bento_model)

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                self.predict_fns[method_name] = getattr(self.model, method_name)

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(self: EasyOCRRunnable, *args: t.Any, **kwargs: t.Any) -> t.Any:
            return self.predict_fns[method_name](*args, **kwargs)

        EasyOCRRunnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return EasyOCRRunnable



from __future__ import annotations

import importlib.metadata
import importlib.util
import logging

from packaging import version

logger = logging.getLogger(__name__)


def is_flax_available():
    _flax_available = (
        importlib.util.find_spec("jax") is not None
        and importlib.util.find_spec("flax") is not None
    )
    if _flax_available:
        try:
            importlib.metadata.version("jax")
            importlib.metadata.version("flax")
        except importlib.metadata.PackageNotFoundError:
            _flax_available = False
    return _flax_available


def is_torch_available():
    _torch_available = importlib.util.find_spec("torch") is not None
    if _torch_available:
        try:
            importlib.metadata.version("torch")
        except importlib.metadata.PackageNotFoundError:
            _torch_available = False
    return _torch_available


def is_tf_available():
    from .tensorflow import get_tf_version

    _tf_available = importlib.util.find_spec("tensorflow") is not None
    if _tf_available:
        _tf_version = get_tf_version()
        _tf_available = _tf_version != ""
        if version.parse(_tf_version) < version.parse("2"):
            logger.info(
                f"Tensorflow found but with verison {_tf_version}. Transformers support with BentoML requires a minimum of Tensorflow 2 and above."
            )
            _tf_available = False
    return _tf_available



from __future__ import annotations

import logging
import os
import shutil
import tempfile
import typing as t
from typing import TYPE_CHECKING

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException
from bentoml.exceptions import NotFound
from bentoml.models import ModelContext

if TYPE_CHECKING:
    from types import ModuleType

    from bentoml.types import ModelSignature
    from bentoml.types import ModelSignatureDict


try:
    import mlflow
    import mlflow.models
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'mlflow' is required in order to use module 'bentoml.mlflow', install mlflow with 'pip install mlflow'. For more information, refer to https://mlflow.org/",
    )


MODULE_NAME = "bentoml.mlflow"
MLFLOW_MODEL_FOLDER = "mlflow_model"
API_VERSION = "v1"

logger = logging.getLogger(__name__)


def get(tag_like: str | Tag) -> bentoml.Model:
    """
    Get the BentoML model with the given tag.

    Args:
        tag_like: The tag of the model to retrieve from the model store.

    Returns:
        :obj:`~bentoml.Model`: A BentoML :obj:`~bentoml.Model` with the matching tag.

    Example:

    .. code-block:: python

       import bentoml
       # target model must be from the BentoML model store
       model = bentoml.mlflow.get("my_mlflow_model")
    """
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def load_model(
    bento_model: str | Tag | bentoml.Model,
) -> mlflow.pyfunc.PyFuncModel:
    """
    Load the MLflow `PyFunc <https://www.mlflow.org/docs/latest/python_api/mlflow.pyfunc.html#mlflow.pyfunc.PyFuncModel>`_ model with the given tag from the local BentoML model store.

    Args:
        bento_model: Either the tag of the model to get from the store, or a BentoML
            ``~bentoml.Model`` instance to load the model from.

    Returns:
        The MLflow model loaded as PyFuncModel from the BentoML model store.

    Example:

    .. code-block:: python

        import bentoml
        pyfunc_model = bentoml.mlflow.load_model('my_model:latest')
        pyfunc_model.predict( input_df )
    """  # noqa
    if not isinstance(bento_model, bentoml.Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    return mlflow.pyfunc.load_model(bento_model.path_of(MLFLOW_MODEL_FOLDER))


def import_model(
    name: Tag | str,
    model_uri: str,
    *,
    signatures: dict[str, ModelSignature] | dict[str, ModelSignatureDict] | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
    # ...
) -> bentoml.Model:
    """
    Import MLflow model from a artifact URI to the BentoML model store.

    Args:
        name:
            The name to give to the model in the BentoML store. This must be a valid
            :obj:`~bentoml.Tag` name.
        model_uri:
            The MLflow model to be saved.
        signatures:
            Signatures of predict methods to be used. If not provided, the signatures
            default to {"predict": {"batchable": False}}. See
            :obj:`~bentoml.types.ModelSignature` for more details.
        labels:
            A default set of management labels to be associated with the model. For
            example: ``{"training-set": "data-v1"}``.
        custom_objects:
            Custom objects to be saved with the model. An example is
            ``{"my-normalizer": normalizer}``. Custom objects are serialized with
            cloudpickle.
        metadata:
            Metadata to be associated with the model. An example is ``{"param_a": .2}``.

            Metadata is intended for display in a model management UI and therefore all
            values in metadata dictionary must be a primitive Python type, such as
            ``str`` or ``int``.

    Returns:
        A :obj:`~bentoml.Model` instance referencing a saved model in the local BentoML
        model store.

    Example:

    .. code-block:: python

        import bentoml

        bentoml.mlflow.import_model(
            'my_mlflow_model',
            model_uri="runs:/<mlflow_run_id>/run-relative/path/to/model",
            signatures={
                "predict": {"batchable": True},
            }
        )
    """
    context = ModelContext(
        framework_name="mlflow",
        framework_versions={"mlflow": mlflow.__version__},
    )

    if signatures is None:
        signatures = {
            "predict": {"batchable": False},
        }
        logger.info(
            'Using the default model signature for MLflow (%s) for model "%s".',
            signatures,
            name,
        )
    if len(signatures) != 1 or "predict" not in signatures:
        raise BentoMLException(
            f"MLflow pyfunc model support only the `predict` method, signatures={signatures} is not supported"
        )

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        signatures=signatures,
        labels=labels,
        options=None,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        context=context,
    ) as bento_model:
        from mlflow.models import Model as MLflowModel
        from mlflow.models.model import MLMODEL_FILE_NAME
        from mlflow.pyfunc import FLAVOR_NAME as PYFUNC_FLAVOR_NAME

        # Explicitly provide a destination dir to mlflow so that we don't
        # accidentially download into the root of the bento model temp dir
        # (using a model:/ url can cause this)
        download_dir = tempfile.mkdtemp(dir=bento_model.path)

        try:
            # Prefer public API download_artifacts introduced in MLflow 1.25
            from mlflow.artifacts import download_artifacts

            local_path = download_artifacts(
                artifact_uri=model_uri, dst_path=download_dir
            )
        except (ModuleNotFoundError, ImportError):
            # For MLflow < 1.25
            from mlflow.tracking.artifact_utils import _download_artifact_from_uri

            local_path: str = _download_artifact_from_uri(
                artifact_uri=model_uri, output_path=download_dir
            )
        finally:
            mlflow_model_path = bento_model.path_of(MLFLOW_MODEL_FOLDER)
            # Rename model folder from original artifact name to fixed "mlflow_model"
            shutil.move(local_path, mlflow_model_path)  # type: ignore (local_path is bound)
            # Remove the tempdir if it still exists.
            # NOTE for models:/ uri downloads, the download_dir itself is actually renamed
            # in the previous line, not a subdir of download_dir like other methods.
            # Calling rmtree unchecked will lead to models:/ downloads failing
            if os.path.exists(download_dir):
                shutil.rmtree(download_dir)

        mlflow_model_file = os.path.join(mlflow_model_path, MLMODEL_FILE_NAME)

        if not os.path.exists(mlflow_model_file):
            raise BentoMLException(f'artifact "{model_uri}" is not a MLflow model')

        model_meta = MLflowModel.load(mlflow_model_file)
        if PYFUNC_FLAVOR_NAME not in model_meta.flavors:
            raise BentoMLException(
                f'MLflow model "{model_uri}" does not support the required python_function flavor'
            )

        return bento_model


def get_runnable(bento_model: bentoml.Model) -> t.Type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """
    assert "predict" in bento_model.info.signatures
    predict_signature = bento_model.info.signatures["predict"]

    class MLflowPyfuncRunnable(bentoml.Runnable):
        # The only case that multi-threading may not be supported is when user define a
        # custom python_function MLflow model with pure python code, but there's no way
        # of telling that from the MLflow model metadata. It should be a very rare case,
        # because most custom python_function models are likely numpy code or model
        # inference with pre/post-processing code.
        SUPPORTED_RESOURCES = ("cpu",)
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()
            self.model = load_model(bento_model)

        @bentoml.Runnable.method(
            batchable=predict_signature.batchable,
            batch_dim=predict_signature.batch_dim,
            input_spec=None,
            output_spec=None,
        )
        def predict(self, input_data: t.Any) -> t.Any:
            return self.model.predict(input_data)

    return MLflowPyfuncRunnable


def get_mlflow_model(tag_like: str | Tag) -> mlflow.models.Model:
    bento_model = get(tag_like)
    return mlflow.models.Model.load(bento_model.path_of(MLFLOW_MODEL_FOLDER))



from __future__ import annotations

import typing as t
from types import ModuleType
from typing import TYPE_CHECKING

import bentoml
from bentoml import Tag

from ...exceptions import MissingDependencyException
from ...exceptions import NotFound
from ..models.model import Model
from .common.pytorch import torch
from .torchscript import MODEL_FILENAME
from .torchscript import save_model as script_save_model

if TYPE_CHECKING:
    from ..models.model import ModelSignaturesType

try:
    import pytorch_lightning as pl
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'pytorch_lightning' is required in order to use module 'bentoml.pytorch_lightning', install python-lightning with: 'pip install pytorch-lightning'"
    )

MODULE_NAME = "bentoml.pytorch_lightning"


__all__ = ["save_model", "load_model", "get_runnable", "get"]


def get(tag_like: str | Tag) -> Model:
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def load_model(
    bentoml_model: str | Tag | Model,
    device_id: t.Optional[str] = "cpu",
) -> torch.ScriptModule:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        device_id (:code:`str`, `optional`):
            Optional devices to put the given model on. Refer to https://pytorch.org/docs/stable/tensor_attributes.html#torch.torch.device
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`torch.ScriptModule`: an instance of :obj:`torch.ScriptModule` from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml
        lit = bentoml.torchscript.load_model('lit_classifier:latest', device_id="cuda:0")
    """
    if isinstance(bentoml_model, (str, Tag)):
        bentoml_model = get(bentoml_model)

    if bentoml_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bentoml_model.tag} was saved with module {bentoml_model.info.module}, not loading with {MODULE_NAME}."
        )
    weight_file = bentoml_model.path_of(MODEL_FILENAME)
    model: torch.ScriptModule = torch.jit.load(weight_file, map_location=device_id)
    return model


def save_model(
    name: Tag | str,
    model: pl.LightningModule,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: t.Dict[str, str] | None = None,
    custom_objects: t.Dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: t.Dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`pl.LightningModule`):
            Instance of model to be saved
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        external_modules (:code:`List[ModuleType]`, `optional`, default to :code:`None`):
            user-defined additional python modules to be saved alongside the model or custom objects,
            e.g. a tokenizer module, preprocessor module, model configuration module
        metadata (:code:`Dict[str, Any]`, `optional`, default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import bentoml
        import torch
        import pytorch_lightning as pl

        class LitClassifier(pl.LightningModule):

            def __init__(self, hidden_dim: int = 128, learning_rate: float = 0.0001):
                super().__init__()
                self.save_hyperparameters()

                self.l1 = torch.nn.Linear(28 * 28, self.hparams.hidden_dim)
                self.l2 = torch.nn.Linear(self.hparams.hidden_dim, 10)

            def forward(self, x):
                x = x.view(x.size(0), -1)
                x = torch.relu(self.l1(x))
                x = torch.relu(self.l2(x))
                return x

            def training_step(self, batch, batch_idx):
                x, y = batch
                y_hat = self(x)
                loss = F.cross_entropy(y_hat, y)
                return loss

            def validation_step(self, batch, batch_idx):
                x, y = batch
                y_hat = self(x)
                loss = F.cross_entropy(y_hat, y)
                self.log("valid_loss", loss)

            def test_step(self, batch, batch_idx):
                x, y = batch
                y_hat = self(x)
                loss = F.cross_entropy(y_hat, y)
                self.log("test_loss", loss)

            def configure_optimizers(self):
                return torch.optim.Adam(self.parameters(), lr=self.hparams.learning_rate)

        tag = bentoml.pytorch_lightning.save_model("lit_classifier", LitClassifier())
    """
    if not isinstance(model, pl.LightningModule):
        raise TypeError(
            f"'model' must be an instance of 'pl.LightningModule', got {type(model)} instead."
        )

    script_module = model.to_torchscript()

    assert not isinstance(
        script_module, dict
    ), "Saving a dict of pytorch_lightning Module into one BentoModel is not supported"

    return script_save_model(
        name,
        script_module,
        signatures=signatures,
        labels=labels,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        _framework_name="pytorch_lightning",
        _module_name=MODULE_NAME,
    )


def get_runnable(bento_model: Model):
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """
    from .common.pytorch import PytorchModelRunnable
    from .common.pytorch import make_pytorch_runnable_method
    from .common.pytorch import partial_class

    for method_name, options in bento_model.info.signatures.items():
        PytorchModelRunnable.add_method(
            make_pytorch_runnable_method(method_name),
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )
    return partial_class(
        PytorchModelRunnable,
        bento_model=bento_model,
        loader=load_model,
    )



from __future__ import annotations

import asyncio
import functools
import logging
import os
import sys
import typing as t

from simple_di import Provide
from simple_di import inject
from starlette.exceptions import HTTPException
from starlette.responses import PlainTextResponse

from ...exceptions import BentoMLException
from ..configuration.containers import BentoMLContainer
from ..context import trace_context
from ..server.base_app import BaseAppFactory
from ..service.service import Service

if t.TYPE_CHECKING:
    from opentelemetry.sdk.trace import Span
    from starlette.applications import Starlette
    from starlette.middleware import Middleware
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.routing import BaseRoute

    from ..service.inference_api import InferenceAPI

    LifecycleHook = t.Callable[[], None | t.Coroutine[t.Any, t.Any, None]]


logger = logging.getLogger(__name__)

DEFAULT_INDEX_HTML = """\
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8">
    <title>BentoML Prediction Service</title>
    <!-- Google Tag Manager -->
    <script>(function(w,d,s,l,i){w[l]=w[l]||[];w[l].push({'gtm.start':
    new Date().getTime(),event:'gtm.js'});var f=d.getElementsByTagName(s)[0],
    j=d.createElement(s),dl=l!='dataLayer'?'&l='+l:'';j.async=true;j.src=
    'https://www.googletagmanager.com/gtm.js?id='+i+dl;f.parentNode.insertBefore(j,f);
    })(window,document,'script','dataLayer','GTM-WNPGWRM');</script>
    <!-- End Google Tag Manager -->
    <link rel="stylesheet" type="text/css" href="./static_content/swagger-ui.css" />
    <link rel="stylesheet" type="text/css" href="./static_content/index.css" />
    <link rel="icon" type="image/png" href="./static_content/favicon-light-32x32.png" sizes="32x32" media="(prefers-color-scheme: light)" />
    <link rel="icon" type="image/png" href="./static_content/favicon-dark-32x32.png" sizes="32x32" media="(prefers-color-scheme: dark)" />
  </head>
  <body>
    <!-- Google Tag Manager (noscript) -->
    <noscript><iframe src="https://www.googletagmanager.com/ns.html?id=GTM-WNPGWRM"
    height="0" width="0" style="display:none;visibility:hidden"></iframe></noscript>
    <!-- End Google Tag Manager (noscript) -->
    <div id="swagger-ui"></div>
    <script src="./static_content/swagger-ui-bundle.js" charset="UTF-8"> </script>
    <script src="./static_content/swagger-ui-standalone-preset.js" charset="UTF-8"> </script>
    <script src="./static_content/swagger-initializer.js" charset="UTF-8"> </script>
    <div class="version">
        <div class="version-section"><a href="https://github.com/bentoml/BentoML" class="github-corner" aria-label="Powered by BentoML">
            <svg width="80" height="80" viewBox="0 0 250 250" style="fill:#151513; color:#fff; position: absolute; top: 0; border: 0; right: 0;" aria-hidden="true">
            <path d="M0,0 L115,115 L130,115 L142,142 L250,250 L250,0 Z"></path>
            <path d="M128.3,109.0 C113.8,99.7 119.0,89.6 119.0,89.6 C122.0,82.7 120.5,78.6 120.5,78.6 C119.2,72.0 123.4,76.3 123.4,76.3 C127.3,80.9 125.5,87.3 125.5,87.3 C122.9,97.6 130.6,101.9 134.4,103.2" fill="currentColor" style="transform-origin: 130px 106px;" class="octo-arm"></path>
            <path d="M115.0,115.0 C114.9,115.1 118.7,116.5 119.8,115.4 L133.7,101.6 C136.9,99.2 139.9,98.4 142.2,98.6 C133.8,88.0 127.5,74.4 143.8,58.0 C148.5,53.4 154.0,51.2 159.7,51.0 C160.3,49.4 163.2,43.6 171.4,40.1 C171.4,40.1 176.1,42.5 178.8,56.2 C183.1,58.6 187.2,61.8 190.9,65.4 C194.5,69.0 197.7,73.2 200.1,77.6 C213.8,80.2 216.3,84.9 216.3,84.9 C212.7,93.1 206.9,96.0 205.4,96.6 C205.1,102.4 203.0,107.8 198.3,112.5 C181.9,128.9 168.3,122.5 157.7,114.1 C157.9,116.9 156.7,120.9 152.7,124.9 L141.0,136.5 C139.8,137.7 141.6,141.9 141.8,141.8 Z" fill="currentColor" class="octo-body"></path>
            </svg></a>
        </div>
    </div>
  </body>
</html>
"""


def log_exception(request: Request, exc_info: t.Any) -> None:
    """
    Logs an exception.  This is called by :meth:`handle_exception`
    if debugging is disabled and right before the handler is called.
    The default implementation logs the exception as error on the
    :attr:`logger`.
    """
    logger.error(
        "Exception on %s [%s]", request.url.path, request.method, exc_info=exc_info
    )


class HTTPAppFactory(BaseAppFactory):
    """
    HTTPApp creates a REST API server based on APIs defined with a BentoService
    via BentoService#apis. Each InferenceAPI will become one
    endpoint exposed on the REST server, and the RequestHandler defined on each
    InferenceAPI object will be used to handle Request object before feeding the
    request data into a Service API function
    """

    @inject
    def __init__(
        self,
        bento_service: Service,
        enable_access_control: bool = Provide[BentoMLContainer.http.cors.enabled],
        access_control_options: dict[str, list[str] | str | int] = Provide[
            BentoMLContainer.access_control_options
        ],
        enable_metrics: bool = Provide[
            BentoMLContainer.api_server_config.metrics.enabled
        ],
        max_concurrency: int
        | None = Provide[BentoMLContainer.api_server_config.traffic.max_concurrency],
    ):
        self.bento_service = bento_service
        self.enable_access_control = enable_access_control
        self.access_control_options = access_control_options
        self.enable_metrics = enable_metrics
        timeout = BentoMLContainer.api_server_config.traffic.timeout.get()
        super().__init__(timeout=timeout, max_concurrency=max_concurrency)

    @property
    def name(self) -> str:
        return self.bento_service.name

    async def index_view_func(self, _: Request) -> Response:
        """
        The default index view for BentoML API server. This includes the readme
        generated from docstring and swagger UI
        """
        from starlette.responses import Response

        # TODO: add readme description.
        return Response(
            content=DEFAULT_INDEX_HTML,
            status_code=200,
            media_type="text/html",
        )

    async def docs_view_func(self, _: Request) -> Response:
        from starlette.responses import JSONResponse

        return JSONResponse(self.bento_service.openapi_spec.asdict())

    @property
    def routes(self) -> list[BaseRoute]:
        """
        Setup routes for bento model server, including:

        /               Index Page, shows readme docs, metadata, and Swagger UI
        /docs.json      Returns Swagger/OpenAPI definition file in json format
        /healthz        liveness probe endpoint
        /readyz         Readiness probe endpoint
        /metrics        Prometheus metrics endpoint

        And user defined InferenceAPI list into routes, e.g.:
        /classify
        /predict
        """
        from starlette.routing import Mount
        from starlette.routing import Route
        from starlette.staticfiles import StaticFiles

        routes = super().routes

        routes.append(Route(path="/", name="home", endpoint=self.index_view_func))
        routes.append(
            Route(
                path="/docs.json",
                name="docs",
                endpoint=self.docs_view_func,
            )
        )

        parent_dir_path = os.path.dirname(os.path.realpath(__file__))
        routes.append(
            Mount(
                "/static_content",
                app=StaticFiles(
                    directory=os.path.join(parent_dir_path, "static_content")
                ),
                name="static_content",
            )
        )

        for _, api in self.bento_service.apis.items():
            api_route_endpoint = self._create_api_endpoint(api)
            if api.route.startswith("/"):
                route_path = api.route
            else:
                route_path = f"/{api.route}"

            routes.append(
                Route(
                    path=route_path,
                    name=api.name,
                    endpoint=api_route_endpoint,
                    methods=api.input.HTTP_METHODS,
                )
            )

        return routes

    @property
    def middlewares(self) -> list[Middleware]:
        middlewares = super().middlewares

        from starlette.middleware import Middleware

        for middleware_cls, options in self.bento_service.middlewares:
            middlewares.append(Middleware(middleware_cls, **options))

        if self.enable_access_control:
            assert (
                self.access_control_options.get("allow_origins") is not None
            ), "To enable cors, access_control_allow_origin must be set"

            from starlette.middleware.cors import CORSMiddleware

            middlewares.append(
                Middleware(CORSMiddleware, **self.access_control_options)
            )

        # metrics middleware
        if self.enable_metrics:
            from .http.instruments import HTTPTrafficMetricsMiddleware

            middlewares.append(Middleware(HTTPTrafficMetricsMiddleware))

        # otel middleware
        from opentelemetry.instrumentation.asgi import OpenTelemetryMiddleware

        def client_request_hook(span: Span, _: dict[str, t.Any]) -> None:
            if span is not None:
                trace_context.request_id = span.context.span_id

        middlewares.append(
            Middleware(
                OpenTelemetryMiddleware,
                excluded_urls=BentoMLContainer.tracing_excluded_urls.get(),
                default_span_details=None,
                server_request_hook=None,
                client_request_hook=client_request_hook,
                tracer_provider=BentoMLContainer.tracer_provider.get(),
            )
        )

        access_log_config = BentoMLContainer.api_server_config.logging.access
        if access_log_config.enabled.get():
            from .http.access import AccessLogMiddleware

            access_logger = logging.getLogger("bentoml.access")
            if access_logger.getEffectiveLevel() <= logging.INFO:
                middlewares.append(
                    Middleware(
                        AccessLogMiddleware,
                        has_request_content_length=access_log_config.request_content_length.get(),
                        has_request_content_type=access_log_config.request_content_type.get(),
                        has_response_content_length=access_log_config.response_content_length.get(),
                        has_response_content_type=access_log_config.response_content_type.get(),
                    )
                )

        return middlewares

    @property
    def on_startup(self) -> list[LifecycleHook]:
        on_startup = [
            *self.bento_service.startup_hooks,
            self.bento_service.on_asgi_app_startup,
        ]
        if BentoMLContainer.development_mode.get():
            for runner in self.bento_service.runners:
                on_startup.append(functools.partial(runner.init_local, quiet=True))
        else:
            for runner in self.bento_service.runners:
                if runner.embedded:
                    on_startup.append(functools.partial(runner.init_local, quiet=True))
                else:
                    on_startup.append(runner.init_client)
        on_startup.extend(super().on_startup)
        return on_startup

    async def readyz(self, _: "Request") -> "Response":
        if BentoMLContainer.api_server_config.runner_probe.enabled.get():
            runner_statuses = (
                runner.runner_handle_is_ready() for runner in self.bento_service.runners
            )
            runners_ready = all(await asyncio.gather(*runner_statuses))

            if not runners_ready:
                raise HTTPException(status_code=503, detail="Runners are not ready.")

        return PlainTextResponse("\n", status_code=200)

    @property
    def on_shutdown(self) -> list[LifecycleHook]:
        on_shutdown = [
            *self.bento_service.shutdown_hooks,
            self.bento_service.on_asgi_app_shutdown,
        ]
        for runner in self.bento_service.runners:
            on_shutdown.append(runner.destroy)
        on_shutdown.extend(super().on_shutdown)
        return on_shutdown

    def __call__(self) -> Starlette:
        app = super().__call__()
        for mount_app, path, name in self.bento_service.mount_apps:
            app.mount(app=mount_app, path=path, name=name)
        return app

    def _create_api_endpoint(
        self,
        api: InferenceAPI,
    ) -> t.Callable[[Request], t.Coroutine[t.Any, t.Any, Response]]:
        """
        Create api function for starlette route, it wraps around user defined API
        callback and adapter class, and adds request logging and instrument metrics
        """
        from starlette.concurrency import run_in_threadpool  # type: ignore
        from starlette.responses import JSONResponse

        from ..utils import is_async_callable

        async def api_func(request: Request) -> Response:
            # handle_request may raise 4xx or 5xx exception.
            output = None
            with self.bento_service.context.in_request(request) as ctx:
                try:
                    input_data = await api.input.from_http_request(request)
                    if api.multi_input:
                        if api.needs_ctx:
                            input_data[api.ctx_param] = ctx
                        if is_async_callable(api.func):
                            output = await api.func(**input_data)
                        else:
                            output = await run_in_threadpool(api.func, **input_data)
                    else:
                        args = (input_data,)
                        if api.needs_ctx:
                            args = (input_data, ctx)
                        if is_async_callable(api.func):
                            output = await api.func(*args)
                        else:
                            output = await run_in_threadpool(api.func, *args)

                    response = await api.output.to_http_response(
                        output, ctx if api.needs_ctx else None
                    )

                    if trace_context.request_id is not None:
                        response.headers["X-BentoML-Request-ID"] = str(
                            trace_context.request_id
                        )
                    if (
                        BentoMLContainer.http.response.trace_id.get()
                        and trace_context.trace_id is not None
                    ):
                        response.headers["X-BentoML-Trace-ID"] = str(
                            trace_context.trace_id
                        )
                except BentoMLException as e:
                    log_exception(request, sys.exc_info())
                    if output is not None:
                        import inspect

                        signature = inspect.signature(api.output.to_proto)
                        param = next(iter(signature.parameters.values()))
                        ann = ""
                        if param is not inspect.Parameter.empty:
                            ann = param.annotation

                        # more descriptive errors if output is available
                        logger.error(
                            "Function '%s' has 'input=%s,output=%s' as IO descriptor, and returns 'result=%s', while expected return type is '%s'",
                            api.name,
                            api.input,
                            api.output,
                            type(output),
                            ann,
                        )

                    status = e.error_code.value
                    if 400 <= status < 500 and status not in (401, 403):
                        response = JSONResponse(
                            content="BentoService error handling API request: %s"
                            % str(e),
                            status_code=status,
                        )
                    else:
                        response = JSONResponse("", status_code=status)
                except Exception:  # pylint: disable=broad-except
                    # For all unexpected error, return 500 by default. For example,
                    # if users' model raises an error of division by zero.
                    log_exception(request, sys.exc_info())

                    response = JSONResponse(
                        "An error has occurred in BentoML user code when handling this request, find the error details in server logs",
                        status_code=500,
                    )
                return response

        return api_func



from __future__ import annotations

import logging
import os
import platform
import typing as t
import warnings
from types import ModuleType

import attr

import bentoml

from ...exceptions import BentoMLException
from ...exceptions import MissingDependencyException
from ...exceptions import NotFound
from ..models.model import Model
from ..models.model import ModelContext
from ..models.model import ModelOptions
from ..models.model import ModelSignature
from ..tag import Tag
from ..types import LazyType
from ..utils import LazyLoader
from ..utils.pkg import get_pkg_version
from ..utils.pkg import pkg_version_info

if t.TYPE_CHECKING:
    import cloudpickle
    import tensorflow as tf
    import torch
    from transformers.models.auto.auto_factory import (
        _BaseAutoModelClass as BaseAutoModelClass,
    )

    from ..models.model import ModelSignaturesType

    class SimpleDefaultMapping(t.TypedDict, total=False):
        pt: tuple[str, ...]
        tf: tuple[str, ...]

    class ModelDefaultMapping(t.TypedDict):
        model: SimpleDefaultMapping

    DefaultMapping = (
        ModelDefaultMapping
        | SimpleDefaultMapping
        | dict[tuple[str, ...], ModelDefaultMapping]
    )

    TupleStr = tuple[str, ...]
    TupleAutoModel = tuple[BaseAutoModelClass, ...]

    class PreTrainedProtocol(t.Protocol):
        @property
        def framework(self) -> str:
            ...

        def save_pretrained(self, save_directory: str, **kwargs: t.Any) -> None:
            ...

        @classmethod
        def from_pretrained(
            cls, pretrained_model_name_or_path: str, *args: t.Any, **kwargs: t.Any
        ) -> PreTrainedProtocol:
            ...

    P = t.ParamSpec("P")

else:
    TupleStr = TupleAutoModel = tuple
    DefaultMapping = SimpleDefaultMapping = ModelDefaultMapping = dict

    torch = LazyLoader("torch", globals(), "torch")
    tf = LazyLoader("tf", globals(), "tensorflow")
    cloudpickle = LazyLoader(
        "cloudpickle",
        globals(),
        "cloudpickle",
        exc_msg="'cloudpickle' is required to save/load custom pipeline.",
    )


try:
    import transformers
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'transformers' is required in order to use module 'bentoml.transformers'. Install transformers with 'pip install transformers'."
    )

if t.TYPE_CHECKING:
    TransformersPreTrained = (
        transformers.PreTrainedTokenizerBase
        | transformers.PreTrainedTokenizer
        | transformers.PreTrainedTokenizerFast
        | transformers.PreTrainedModel
        | transformers.TFPreTrainedModel
        | transformers.FlaxPreTrainedModel
        | transformers.image_processing_utils.BaseImageProcessor
        | transformers.SequenceFeatureExtractor
        | transformers.Pipeline
    )


__all__ = ["load_model", "save_model", "get_runnable", "get"]

_object_setattr = object.__setattr__

MODULE_NAME = "bentoml.transformers"
API_VERSION = "v2"
PIPELINE_PICKLE_NAME = f"pipeline.{API_VERSION}.pkl"
PRETRAINED_PROTOCOL_NAME = f"pretrained.{API_VERSION}.pkl"


logger = logging.getLogger(__name__)

TRANSFORMERS_VERSION = pkg_version_info("transformers")

HAS_PIPELINE_REGISTRY = TRANSFORMERS_VERSION >= (4, 21, 0)


def _deep_convert_to_tuple(
    dct: dict[str, str | TupleStr | list[str] | dict[str, t.Any]]
) -> dict[str, str | TupleStr | list[str] | dict[str, t.Any]]:
    for k, v in dct.items():
        override = v
        if isinstance(v, list):
            override = tuple(v)
        elif isinstance(v, dict):
            override = _deep_convert_to_tuple(v)
        dct[k] = override
    return dct


def _validate_pipeline_type(
    _: t.Any, attribute: attr.Attribute[str | list[str]], value: str | list[str]
) -> None:
    """
    Validate the type of the given pipeline definition. The value is expected to be a `str`.
    `list` type is also allowed here to maintain compatibility with an earlier introduced bug.

    TODO: disallow list type in the next minor version release.
    """
    if not isinstance(value, str) and not isinstance(value, list):
        raise ValueError(f"{attribute.name} must be a string")


if t.TYPE_CHECKING:

    class TaskDefinition(t.TypedDict):
        impl: type[transformers.Pipeline]
        tf: TupleAutoModel | BaseAutoModelClass | None
        pt: TupleAutoModel | BaseAutoModelClass | None
        default: t.NotRequired[DefaultMapping]
        type: t.NotRequired[str]

else:
    TaskDefinition = dict


def _autoclass_converter(
    value: tuple[BaseAutoModelClass | str, ...] | None
) -> TupleStr:
    if value is None:
        return TupleStr()
    elif isinstance(value, t.Iterable):
        value = tuple(value)
    elif not isinstance(value, tuple):
        value = (value,)
    return tuple(it if isinstance(it, str) else it.__qualname__ for it in value)


@attr.define
class TransformersOptions(ModelOptions):
    """Options for the Transformers model."""

    task: str = attr.field(factory=str, validator=attr.validators.instance_of(str))
    tf: TupleStr = attr.field(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(str), attr.validators.instance_of(TupleStr)
        ),
        factory=TupleStr,
        converter=_autoclass_converter,
    )
    pt: TupleStr = attr.field(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(str), attr.validators.instance_of(TupleStr)
        ),
        factory=TupleStr,
        converter=_autoclass_converter,
    )
    default: DefaultMapping = attr.field(factory=dict, converter=_deep_convert_to_tuple)
    type: str = attr.field(
        validator=attr.validators.optional(_validate_pipeline_type),
        default=None,
    )
    kwargs: t.Dict[str, t.Any] = attr.field(factory=dict)

    @staticmethod
    def process_task_mapping(
        impl: type[transformers.Pipeline],
        pt: tuple[BaseAutoModelClass | str, ...]
        | TupleAutoModel
        | BaseAutoModelClass
        | None = None,
        tf: tuple[BaseAutoModelClass | str, ...]
        | TupleAutoModel
        | BaseAutoModelClass
        | None = None,
        default: DefaultMapping | None = None,
        type: str | None = None,
    ) -> TaskDefinition:
        if pt is None:
            pt = TupleAutoModel()
        elif not isinstance(pt, tuple):
            pt = (pt,)
        pt = tuple(convert_to_autoclass(it) if isinstance(it, str) else it for it in pt)

        if tf is None:
            tf = TupleAutoModel()
        elif not isinstance(tf, tuple):
            tf = (tf,)
        tf = tuple(convert_to_autoclass(it) if isinstance(it, str) else it for it in tf)

        task_impl = TaskDefinition(impl=impl, pt=pt, tf=tf)

        if default is not None:
            if "model" not in default and ("pt" in default or "tf" in default):
                # case for SimpleDefaultMapping, needs then to convert to ModelDefaultMapping
                default = ModelDefaultMapping(
                    model=t.cast(SimpleDefaultMapping, default)
                )
            task_impl["default"] = default
        if type is not None:
            task_impl["type"] = type

        return task_impl

    @classmethod
    def from_task(cls, task: str, definition: TaskDefinition) -> TransformersOptions:
        return cls(
            task=task,
            tf=definition.get("tf", None),  # type: ignore (handle by cattrs converter)
            pt=definition.get("pt", None),  # type: ignore (handle by cattrs converter)
            default=definition.get("default", {}),
            type=definition.get("type", "text"),
        )

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "task": self.task,
            "tf": tuple(self.tf),
            "pt": tuple(self.pt),
            "default": {
                "_".join(k) if isinstance(k, tuple) else k: v
                for k, v in self.default.items()
            },
            "type": self.type,
            "kwargs": self.kwargs,
        }


def convert_to_autoclass(cls_name: str) -> BaseAutoModelClass:
    if not hasattr(transformers, cls_name):
        raise BentoMLException(
            f"Given {cls_name} is not a valid Transformers auto class. For more information, "
            "please see https://huggingface.co/docs/transformers/main/en/model_doc/auto"
        )
    return getattr(transformers, cls_name)


def get(tag_like: str | Tag) -> Model:
    """
    Get the BentoML model with the given tag.

    Args:
        tag_like: The tag of the model to retrieve from the model store.

    Returns:
        :obj:`~bentoml.Model`: A BentoML :obj:`~bentoml.Model` with the matching tag.

    .. note::


    Example:

    .. code-block:: python

       import bentoml
       # target model must be from the BentoML model store
       model = bentoml.transformers.get("my_pipeline:latest")
    """
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def register_pipeline(
    task: str,
    impl: type[transformers.Pipeline],
    pt: tuple[BaseAutoModelClass | str, ...]
    | TupleAutoModel
    | BaseAutoModelClass
    | None = None,
    tf: tuple[BaseAutoModelClass | str, ...]
    | TupleAutoModel
    | BaseAutoModelClass
    | None = None,
    default: DefaultMapping | None = None,
    type: str | None = None,
):
    task_impl = TransformersOptions.process_task_mapping(impl, pt, tf, default, type)

    if HAS_PIPELINE_REGISTRY:
        from transformers.pipelines import PIPELINE_REGISTRY

        PIPELINE_REGISTRY.register_pipeline(
            task,
            pipeline_class=impl,
            pt_model=task_impl["pt"],
            tf_model=task_impl["tf"],
            default=task_impl.get("default", None),
            type=task_impl.get("type", None),
        )
    else:
        # For backward compatibility
        from transformers.pipelines import SUPPORTED_TASKS

        SUPPORTED_TASKS.setdefault(task, task_impl)

        _object_setattr(impl, "_registered_impl", {task: task_impl})


def delete_pipeline(task: str) -> None:
    """
    Remove pipelines from current registry by task name.
    """
    if HAS_PIPELINE_REGISTRY:
        from transformers.pipelines import PIPELINE_REGISTRY

        del PIPELINE_REGISTRY.supported_tasks[task]
    else:
        # For backward compatibility
        from transformers.pipelines import SUPPORTED_TASKS

        del SUPPORTED_TASKS[task]


@t.overload
def load_model(
    bento_model: str | Tag | Model, **kwargs: t.Any
) -> transformers.Pipeline:
    ...


@t.overload
def load_model(
    bento_model: str | Tag | Model, *args: t.Any, **kwargs: t.Any
) -> TransformersPreTrained:
    ...


def load_model(bento_model: str | Tag | Model, *args: t.Any, **kwargs: t.Any) -> t.Any:
    """
    Load the Transformers model from BentoML local modelstore with given name.

    Args:
        bento_model: Either the tag of the model to get from the store,
                     or a BentoML :class:`~bentoml.Model` instance to load the
                     model from.
        args: Additional model args to be passed into the model if the object is a Transformers PreTrained protocol.
              This shouldn't be used when the bento_model is a pipeline.
        kwargs: Additional keyword arguments to pass into the pipeline.

    Returns:
        ``transformers.pipeline.base.Pipeline``: The Transformers pipeline loaded
        from the model store.

    Example:

    .. code-block:: python

        import bentoml
        pipeline = bentoml.transformers.load_model('my_model:latest')
    """
    if not isinstance(bento_model, Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    options = t.cast(TransformersOptions, bento_model.info.options)
    api_version = bento_model.info.api_version
    task = options.task
    pipeline: transformers.Pipeline | None = None

    if api_version == "v1":
        # NOTE: backward compatibility
        from transformers.pipelines import get_supported_tasks

        logger.warning(
            "Upgrading '%s' to with newer 'save_model' is recommended. Loading model with version 1.",
            bento_model,
        )

        # Set trust_remote_code to True to allow loading custom pipeline.
        kwargs.setdefault("trust_remote_code", False)
        kwargs.update(options.kwargs)
        if len(kwargs) > 0:
            logger.debug(
                "Loading '%s' pipeline (tag='%s') with kwargs %s.",
                task,
                bento_model.tag,
                kwargs,
            )

        # v1 pickled the pipeline with 'pipeline.v1.pkl'.
        fname = "pipeline.v1.pkl"
        if os.path.exists(bento_model.path_of(fname)):
            with open(bento_model.path_of(fname), "rb") as f:
                pipeline = cloudpickle.load(f)

        if task not in get_supported_tasks():
            logger.debug("'%s' is a custom task. Registering it to the registry.", task)
            kwargs["trust_remote_code"] = True
            # most likely a custom pipeline
            if pipeline is None:
                # if the pipeline is not loaded from the pickle, and it is custom, then we should be able to load
                # it with model=bento_model.path
                return transformers.pipeline(model=bento_model.path, **kwargs)
            else:
                register_pipeline(
                    task,
                    pipeline.__class__,
                    tuple(
                        convert_to_autoclass(auto_class) for auto_class in options.pt
                    ),
                    tuple(
                        convert_to_autoclass(auto_class) for auto_class in options.tf
                    ),
                    options.default,
                    options.type,
                )

        kwargs.setdefault("pipeline_class", pipeline.__class__ if pipeline else None)

        assert (
            task in get_supported_tasks()
        ), f"Task '{task}' is not a valid task for pipeline (available: {get_supported_tasks()})."

        return (
            transformers.pipeline(task=task, model=bento_model.path, **kwargs)
            if pipeline is None
            else pipeline
        )
    else:
        if api_version != "v2":
            logger.warning(
                "Got unknown API version '%s', unexpected errors may occur.",
                api_version,
            )

        if "_pretrained_class" in bento_model.info.metadata:
            with open(bento_model.path_of(PRETRAINED_PROTOCOL_NAME), "rb") as f:
                protocol: PreTrainedProtocol = cloudpickle.load(f)
            return protocol.from_pretrained(bento_model.path, *args, **kwargs)
        else:
            assert (
                len(args) == 0
            ), "Positional args are not supported for pipeline. Make sure to only use kwargs instead."
            with open(bento_model.path_of(PIPELINE_PICKLE_NAME), "rb") as f:
                pipeline_class: type[transformers.Pipeline] = cloudpickle.load(f)

            from transformers.pipelines import get_supported_tasks

            if task not in get_supported_tasks():
                logger.debug(
                    "'%s' is not a supported task, trying to load custom pipeline.",
                    task,
                )

                register_pipeline(
                    task,
                    pipeline_class,
                    tuple(
                        convert_to_autoclass(auto_class) for auto_class in options.pt
                    ),
                    tuple(
                        convert_to_autoclass(auto_class) for auto_class in options.tf
                    ),
                    options.default,
                    options.type,
                )
                kwargs["trust_remote_code"] = True

            kwargs.setdefault("pipeline_class", pipeline_class)

            assert (
                task in get_supported_tasks()
            ), f"Task '{task}' is not a valid task for pipeline (available: {get_supported_tasks()})."

            kwargs.update(options.kwargs)
            if len(kwargs) > 0:
                logger.debug(
                    "Loading '%s' pipeline (tag='%s') with kwargs %s.",
                    task,
                    bento_model.tag,
                    kwargs,
                )
            try:
                return transformers.pipeline(
                    task=task, model=bento_model.path, **kwargs
                )
            except Exception:
                # NOTE: When loading a custom pipeline that is not available on huggingface hub,
                # the class registered in the pipeline registry will be a path to a Python file path.
                # Currently, it doesn't handle relative imports correctly, so users will need to use
                # external_modules when using 'save_model'.
                logger.debug(
                    "If you are loading a custom pipeline, See https://huggingface.co/docs/transformers/main/en/add_new_pipeline#how-to-create-a-custom-pipeline for more information. We recommend to upload the custom pipeline to HuggingFace Hub to ensure consistency. You can also try adding the pipeline instance to 'external_modules': 'import importlib; bentoml.transformers.save_model(..., external_modules=[importlib.import_module(pipeline_instance.__module__)])'"
                )
                raise


def make_default_signatures(pretrained: t.Any) -> ModelSignaturesType:
    default_config = ModelSignature(batchable=False)
    infer_fn = ("__call__",)

    # NOTE: for all processor type recommend to use custom signatures since it is
    # a per case basis.
    if pkg_version_info("transformers")[:2] < (4, 17):
        logger.warning(
            "Given transformers version is less than 4.17.0, signatures inference will be disabled. Make sure to specify the signatures manually."
        )
        return {}

    if transformers.processing_utils.ProcessorMixin in pretrained.__class__.__bases__:
        logger.info(
            "Given '%s' extends the 'transformers.ProcessorMixin'. Make sure to specify the signatures manually if it has additional functions.",
            pretrained.__class__.__name__,
        )
        return {k: default_config for k in ("__call__", "batch_decode", "decode")}

    if isinstance(pretrained, transformers.PreTrainedTokenizerBase):
        infer_fn = (
            "__call__",
            "tokenize",
            "encode",
            "encode_plus",
            "batch_encode_plus",
            "pad",
            "create_token_type_ids_from_sequences",
            "build_inputs_with_special_tokens",
            "prepare_for_model",
            "truncate_sequences",
            "convert_tokens_to_string",
            "batch_decode",
            "decode",
            "get_special_tokens_mask",
            "clean_up_tokenization",
            "prepare_seq2seq_batch",
        )
    elif isinstance(pretrained, transformers.PreTrainedModel):
        infer_fn = (
            "__call__",
            "forward",
            "generate",
            "contrastive_search",
            "greedy_search",
            "sample",
            "beam_search",
            "beam_sample",
            "group_beam_search",
            "constrained_beam_search",
        )
    elif isinstance(pretrained, transformers.TFPreTrainedModel):
        infer_fn = (
            "__call__",
            "predict",
            "call",
            "generate",
            "compute_transition_scores",
            "greedy_search",
            "sample",
            "beam_search",
            "contrastive_search",
        )
    elif isinstance(pretrained, transformers.FlaxPreTrainedModel):
        infer_fn = ("__call__", "generate")
    elif isinstance(pretrained, transformers.image_processing_utils.BaseImageProcessor):
        infer_fn = ("__call__", "preprocess")
    elif isinstance(pretrained, transformers.SequenceFeatureExtractor):
        infer_fn = ("pad",)
    elif not isinstance(pretrained, transformers.Pipeline):
        logger.warning(
            "Unable to infer default signatures for '%s'. Make sure to specify it manually.",
            pretrained,
        )
        return {}

    return {k: default_config for k in infer_fn}


def save_model(
    name: Tag | str,
    pretrained_or_pipeline: TransformersPreTrained
    | transformers.Pipeline
    | PreTrainedProtocol
    | None = None,
    pipeline: transformers.Pipeline | None = None,
    task_name: str | None = None,
    task_definition: dict[str, t.Any] | TaskDefinition | None = None,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
    **save_kwargs: t.Any,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name: Name for given model instance. This should pass Python identifier check.
        pretrained_or_pipeline: Instance of the Transformers pipeline to be saved, or any instance that ``transformers`` supports.
                                The object instance should have ``save_pretrained`` and ``from_pretrained`` (follows the protocol that is defined by `transformers`.)
        task_name: Name of pipeline task. If not provided, the task name will be derived from ``pipeline.task``. Only needed when the target is a custom pipeline.
        task_definition: Task definition for the Transformers custom pipeline. The definition is a dictionary
                         consisting of the following keys:

                        - ``impl`` (``type[transformers.Pipeline]``): The name of the pipeline implementation module. The name should be the same as the pipeline passed in the ``pipeline`` argument.
                        - ``tf`` (:code:`tuple[AnyType]`): The name of the Tensorflow auto model class. One of ``tf`` and ``pt`` auto model class argument is required.
                        - ``pt`` (:code:`tuple[AnyType]`): The name of the PyTorch auto model class. One of ``tf`` and ``pt`` auto model class argument is required.
                        - ``default`` (:code:`Dict[str, AnyType]`): The names of the default models, tokenizers, feature extractors, etc.
                        - ``type`` (:code:`str`): The type of the pipeline, e.g. ``text``, ``audio``, ``image``, ``multimodal``.

                        Example:

                        .. code-block:: python

                            task_definition = {
                                "impl": Text2TextGenerationPipeline,
                                "tf": (TFAutoModelForSeq2SeqLM,) if is_tf_available() else (),
                                "pt": (AutoModelForSeq2SeqLM,) if is_torch_available() else (),
                                "default": {"model": {"pt": "t5-base", "tf": "t5-base"}},
                                "type": "text",
                            }

                        See module `src/transformers/pipelines/__init__.py <https://github.com/huggingface/transformers/blob/main/src/transformers/pipelines/__init__.py#L129>`_ for more details.
        signatures: Methods to expose for running inference on the target model. Signatures are used for creating :obj:`~bentoml.Runner` instances when serving model with :obj:`~bentoml.Service`
        labels: User-defined labels for managing models, e.g. ``team=nlp``, ``stage=dev``.
        custom_objects: Custom objects to be saved with the model. An example is ``{"my-normalizer": normalizer}``.

                        Custom objects are currently serialized with cloudpickle, but this implementation is subject to change.
        external_modules: user-defined additional python modules to be saved alongside the model or custom objects,
                          e.g. a tokenizer module, preprocessor module, model configuration module
        metadata: Custom metadata for given model.
        save_kwargs: Additional keyword arguments to be pass to ````

    .. note::

        Both arguments ``task_name`` and ``task_definition`` must be provided to set save a custom pipeline.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format ``name:version`` where ``name`` is the user-defined model's name, and a generated ``version``.

    Examples:

    .. code-block:: python

        import bentoml

        from transformers import pipeline, AutoTokenizer, AutoModelForCausalLM

        tokenizer = AutoTokenizer.from_pretrained("distilgpt2")
        model = AutoModelForCausalLM.from_pretrained("distilgpt2")
        generator = pipeline(task="text-generation", model=model, tokenizer=tokenizer)
        bento_model = bentoml.transformers.save_model("text-generation-pipeline", generator)
    """  # noqa
    # backward compatibility
    if pipeline is not None:
        warnings.warn(
            f"The 'pipeline={pipeline}' argument is deprecated and will be removed in the future. Please use 'pretrained_or_pipeline' instead.",
            DeprecationWarning,
        )
        pretrained_or_pipeline = pipeline

    assert (
        pretrained_or_pipeline is not None
    ), "Please provide a pipeline or a pretrained object as a second argument."

    # The below API are introduced since 4.18
    if pkg_version_info("transformers")[:2] >= (4, 18):
        from transformers.utils import is_flax_available
        from transformers.utils import is_tf_available
        from transformers.utils import is_torch_available
    else:
        from .utils.transformers import is_flax_available
        from .utils.transformers import is_tf_available
        from .utils.transformers import is_torch_available

    framework_versions = {"transformers": get_pkg_version("transformers")}
    if is_torch_available():
        framework_versions["torch"] = get_pkg_version("torch")
    if is_tf_available():
        from .utils.tensorflow import get_tf_version

        framework_versions[
            "tensorflow-macos" if platform.system() == "Darwin" else "tensorflow"
        ] = get_tf_version()
    if is_flax_available():
        framework_versions.update(
            {
                "flax": get_pkg_version("flax"),
                "jax": get_pkg_version("jax"),
                "jaxlib": get_pkg_version("jaxlib"),
            }
        )
    context = ModelContext(
        framework_name="transformers", framework_versions=framework_versions
    )

    if signatures is None:
        signatures = make_default_signatures(pretrained_or_pipeline)
        # NOTE: ``make_default_signatures`` can return an empty dict, hence we will only
        # log when signatures are available.
        if signatures:
            logger.info(
                'Using the default model signature for Transformers (%s) for model "%s".',
                signatures,
                name,
            )

    if LazyType("transformers.Pipeline").isinstance(pretrained_or_pipeline):
        from transformers.pipelines import check_task
        from transformers.pipelines import get_supported_tasks

        # NOTE: safe casting to annotate task_definition types
        task_definition = (
            t.cast(TaskDefinition, task_definition)
            if task_definition is not None
            else task_definition
        )

        pipeline_ = t.cast("transformers.Pipeline", pretrained_or_pipeline)

        if task_name is not None and task_definition is not None:
            logger.info(
                "Arguments 'task_name' and 'task_definition' are provided. Saving model with pipeline task name '%s' and task definition '%s'.",
                task_name,
                task_definition,
            )
            if pipeline_.task != task_name:
                raise BentoMLException(
                    f"Argument 'task_name' '{task_name}' does not match pipeline task name '{pipeline_.task}'."
                )

            assert "impl" in task_definition, "'task_definition' requires 'impl' key."

            impl = task_definition["impl"]
            if type(pipeline_) != impl:
                raise BentoMLException(
                    f"Argument 'pipeline' is not an instance of {impl}. It is an instance of {type(pipeline_)}."
                )
            options_args = (task_name, task_definition)

            if task_name not in get_supported_tasks():
                register_pipeline(task_name, **task_definition)
                logger.info(
                    "Task '%s' is a custom task and has been registered to the pipeline registry.",
                    task_name,
                )

            assert (
                task_name in get_supported_tasks()
            ), f"Task '{task_name}' failed to register into pipeline registry."
        else:
            assert (
                task_definition is None
            ), "'task_definition' must be None if 'task_name' is not provided."

            # if task_name is None, then we derive the task from pipeline.task
            options_args = t.cast(
                "tuple[str, TaskDefinition]",
                check_task(pipeline_.task if task_name is None else task_name)[:2],
            )

        with bentoml.models.create(
            name,
            module=MODULE_NAME,
            api_version=API_VERSION,
            labels=labels,
            context=context,
            options=TransformersOptions.from_task(*options_args),
            signatures=signatures,
            custom_objects=custom_objects,
            external_modules=external_modules,
            metadata=metadata,
        ) as bento_model:
            pipeline_.save_pretrained(bento_model.path, **save_kwargs)

            # NOTE: we want to pickle the class so that tensorflow, flax pipeline will also work.
            # the weights is already save, so we only need to save the class.
            with open(bento_model.path_of(PIPELINE_PICKLE_NAME), "wb") as f:
                cloudpickle.dump(pipeline_.__class__, f)
            return bento_model
    else:
        pretrained = t.cast("PreTrainedProtocol", pretrained_or_pipeline)
        assert all(
            hasattr(pretrained, defn) for defn in ("save_pretrained", "from_pretrained")
        ), f"'pretrained={pretrained}' is not a valid Transformers object. It must have 'save_pretrained' and 'from_pretrained' methods."
        if metadata is None:
            metadata = {}

        metadata.update(
            {
                "_pretrained_class": pretrained.__class__.__name__,
            }
        )
        if hasattr(pretrained, "framework") and isinstance(
            pretrained,
            (
                transformers.PreTrainedModel,
                transformers.TFPreTrainedModel,
                transformers.FlaxPreTrainedModel,
            ),
        ):
            # NOTE: Only PreTrainedModel and variants has this, not tokenizer.
            metadata["_framework"] = pretrained.framework

        with bentoml.models.create(
            name,
            module=MODULE_NAME,
            api_version=API_VERSION,
            labels=labels,
            context=context,
            options=TransformersOptions(),
            signatures=signatures,
            custom_objects=custom_objects,
            external_modules=external_modules,
            metadata=metadata,
        ) as bento_model:
            pretrained.save_pretrained(bento_model.path, **save_kwargs)

            with open(bento_model.path_of(PRETRAINED_PROTOCOL_NAME), "wb") as f:
                cloudpickle.dump(pretrained.__class__, f)
            return bento_model


def get_runnable(bento_model: bentoml.Model) -> type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class TransformersRunnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()

            available_gpus = os.getenv("CUDA_VISIBLE_DEVICES", "")
            # assign CPU resources
            kwargs = {}

            if available_gpus not in ("", "-1"):
                # assign GPU resources
                if not available_gpus.isdigit():
                    raise ValueError(
                        f"Expecting numeric value for CUDA_VISIBLE_DEVICES, got {available_gpus}."
                    )
                if "_pretrained_class" not in bento_model.info.metadata:
                    # NOTE: then this is a pipeline. We then pass the device to it.
                    kwargs["device"] = int(available_gpus)
            if "_pretrained_class" not in bento_model.info.metadata:
                self.model = load_model(bento_model, **kwargs)
            else:
                if "_framework" in bento_model.info.metadata:
                    if "torch" == bento_model.info.metadata["_framework"]:
                        self.model = t.cast(
                            transformers.PreTrainedModel,
                            load_model(bento_model, **kwargs),
                        ).to(
                            torch.device(
                                "cuda" if available_gpus not in ("", "-1") else "cpu"
                            )
                        )
                        torch.set_default_tensor_type("torch.cuda.FloatTensor")
                    elif "tf" == bento_model.info.metadata["_framework"]:
                        with tf.device(
                            "/device:CPU:0"
                            if available_gpus in ("", "-1")
                            else f"/device:GPU:{available_gpus}"
                        ):
                            self.model = t.cast(
                                transformers.TFPreTrainedModel,
                                load_model(bento_model, **kwargs),
                            )
                    else:
                        # NOTE: we need to hide all GPU from TensorFlow, otherwise it will try to allocate
                        # memory on the GPU and make it unavailable for JAX.
                        tf.config.experimental.set_visible_devices([], "GPU")
                        self.model = t.cast(
                            transformers.FlaxPreTrainedModel,
                            load_model(bento_model, **kwargs),
                        )
                else:
                    logger.warning(
                        "Current %s is saved with an older version of BentoML. Setting GPUs on this won't work as expected. Make sure to save it with a newer version of BentoML.",
                        bento_model,
                    )
                    self.model = load_model(bento_model, **kwargs)

            # NOTE: backward compatibility with previous BentoML versions.
            self.pipeline = self.model

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                self.predict_fns[method_name] = getattr(self.model, method_name)

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(self: TransformersRunnable, *args: t.Any, **kwargs: t.Any) -> t.Any:
            return self.predict_fns[method_name](*args, **kwargs)

        TransformersRunnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return TransformersRunnable



from __future__ import annotations

import contextlib
import itertools
import logging
import pickle
import typing as t
from types import ModuleType
from typing import TYPE_CHECKING

import bentoml
from bentoml import Runnable
from bentoml import Tag
from bentoml.exceptions import MissingDependencyException
from bentoml.exceptions import NotFound
from bentoml.models import ModelContext

from ..models.model import ModelSignature
from ..models.model import PartialKwargsModelOptions as ModelOptions
from ..runner.container import DataContainer
from ..runner.container import DataContainerRegistry
from ..runner.container import Payload
from ..types import LazyType
from .utils.tensorflow import cast_py_args_to_tf_function_args
from .utils.tensorflow import get_input_signatures_v2
from .utils.tensorflow import get_output_signatures_v2
from .utils.tensorflow import get_restorable_functions
from .utils.tensorflow import get_tf_version

if TYPE_CHECKING:
    from .. import external_typing as ext
    from ..external_typing import tensorflow as tf_ext
    from ..models.model import ModelSignatureDict

    TFArgType = t.Union[t.List[t.Union[int, float]], ext.NpNDArray, tf_ext.Tensor]
    TFModelOutputType = tf_ext.EagerTensor | tuple[tf_ext.EagerTensor]
    TFRunnableOutputType = ext.NpNDArray | tuple[ext.NpNDArray]


try:
    import tensorflow as tf
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'tensorflow' is required in order to use module 'bentoml.tensorflow', install tensorflow with 'pip install tensorflow'. For more information, refer to https://www.tensorflow.org/install"
    )

MODULE_NAME = "bentoml.tensorflow"
API_VERSION = "v1"

logger = logging.getLogger(__name__)


def get(tag_like: str | Tag) -> bentoml.Model:
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def load_model(
    bento_model: str | Tag | bentoml.Model,
    device_name: str = "/device:CPU:0",
) -> tf_ext.AutoTrackable | tf_ext.Module:
    """
    Load a tensorflow model from BentoML local modelstore with given name.

    Args:
        bento_model: Either the tag of the model to get from the store, or a
                     BentoML `~bentoml.Model` instance to load the model from.
        device_name: The device id to load the model on. The device id format
                     should be compatible with `tf.device <https://www.tensorflow.org/api_docs/python/tf/device>`_

    Returns:
        :obj:`SavedModel`: an instance of :obj:`SavedModel` format from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml

        # load a model back into memory
        model = bentoml.tensorflow.load_model("my_tensorflow_model")

    """  # noqa
    if not isinstance(bento_model, bentoml.Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    if "GPU" in device_name:
        physical_devices = tf.config.list_physical_devices("GPU")
        try:
            # an optimization for GPU memory growth. But it will raise an error if any
            # tensorflow session is already created. That happens when users test runners
            # in a notebook or Python interactive shell. Thus we just ignore the error.
            tf.config.experimental.set_memory_growth(physical_devices[0], True)
        except RuntimeError:
            pass

    with tf.device(device_name):  # type: ignore (tf.device is a context manager)
        tf_model: tf_ext.AutoTrackable = tf.saved_model.load(bento_model.path)
        return tf_model


def save_model(
    name: Tag | str,
    model: tf_ext.KerasModel | tf_ext.Module,
    *,
    tf_signatures: tf_ext.ConcreteFunction | None = None,
    tf_save_options: tf_ext.SaveOptions | None = None,
    signatures: dict[str, ModelSignature] | dict[str, ModelSignatureDict] | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: list[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name: Name for given model instance. This should pass Python identifier check.
        model: Instance of model to be saved
        tf_signatures: Refer to `Signatures explanation <https://www.tensorflow.org/api_docs/python/tf/saved_model/save>`_
                       from Tensorflow documentation for more information.
        tf_save_options: TensorFlow save options..
        signatures: Methods to expose for running inference on the target model. Signatures are
                    used for creating Runner instances when serving model with bentoml.Service
        labels: user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects: user-defined additional python objects to be saved alongside the model,
                        e.g. a tokenizer instance, preprocessor function, model configuration json
        external_modules: user-defined additional python modules to be saved alongside the model or custom objects,
                          e.g. a tokenizer module, preprocessor module, model configuration module
        metadata: Custom metadata for given model.

    Raises:
        ValueError: If :obj:`obj` is not trackable.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format ``name:version`` where ``name`` is
        the user-defined model's name, and a generated ``version`` by BentoML.

    Examples:

    .. code-block:: python

        import tensorflow as tf
        import numpy as np
        import bentoml

        class NativeModel(tf.Module):
            def __init__(self):
                super().__init__()
                self.weights = np.asfarray([[1.0], [1.0], [1.0], [1.0], [1.0]])
                self.dense = lambda inputs: tf.matmul(inputs, self.weights)

            @tf.function(
                input_signature=[tf.TensorSpec(shape=[1, 5], dtype=tf.float64, name="inputs")]
            )
            def __call__(self, inputs):
                return self.dense(inputs)

        # then save the given model to BentoML modelstore:
        model = NativeModel()
        bento_model = bentoml.tensorflow.save_model("native_toy", model)

    .. note::

       :code:`bentoml.tensorflow.save_model` API also support saving `RaggedTensor <https://www.tensorflow.org/guide/ragged_tensor>`_ model and Keras model. If you choose to save a Keras model
       with :code:`bentoml.tensorflow.save_model`, then the model will be saved under a :obj:`SavedModel` format instead of :obj:`.h5`.

    """  # noqa
    context = ModelContext(
        framework_name="tensorflow",
        framework_versions={"tensorflow": get_tf_version()},
    )

    if signatures is None:
        restorable_functions = get_restorable_functions(model)
        if restorable_functions:
            signatures = {
                k: {
                    "batchable": False,
                }
                for k in restorable_functions
            }
        else:
            signatures = {"__call__": {"batchable": False}}
        logger.info(
            'Using the default model signature for Tensorflow (%s) for model "%s".',
            signatures,
            name,
        )

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        options=ModelOptions(),
        context=context,
        labels=labels,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        signatures=signatures,  # type: ignore
    ) as bento_model:
        tf.saved_model.save(
            model,
            bento_model.path,
            signatures=tf_signatures,
            options=tf_save_options,
        )

        return bento_model


def get_runnable(
    bento_model: bentoml.Model,
):
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    partial_kwargs: t.Dict[str, t.Any] = bento_model.info.options.partial_kwargs

    class TensorflowRunnable(Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()
            if len(tf.config.list_physical_devices("GPU")) > 0:
                # In Multi-GPU scenarios, the visible cuda devices will be set for each Runner worker
                # by the runner's Scheduling Strategy. So that the Runnable implementation only needs
                # to find the first GPU device visible to current process.
                self.device_name = "/device:GPU:0"
            else:
                self.device_name = "/device:CPU:0"

            self.model = load_model(bento_model, device_name=self.device_name)
            self.methods_cache: t.Dict[str, t.Callable[..., t.Any]] = {}
            self.session_stack = contextlib.ExitStack()
            self.session_stack.enter_context(tf.device(self.device_name))

        def __del__(self):
            try:
                self.session_stack.close()
            except RuntimeError:
                pass

    def _gen_run_method(runnable_self: TensorflowRunnable, method_name: str):
        raw_method = getattr(runnable_self.model, method_name)
        method_partial_kwargs = partial_kwargs.get(method_name)

        output_sigs = get_output_signatures_v2(raw_method)

        if len(output_sigs) == 1:
            # if there's only one output signatures, then we can
            # define the _postprocess function without doing
            # conditional casting each time

            sig = output_sigs[0]
            if isinstance(sig, (list, tuple)):

                def _postprocess(
                    res: tuple[tf_ext.EagerTensor],
                ) -> TFRunnableOutputType:
                    return tuple(t.cast("ext.NpNDArray", r.numpy()) for r in res)

            else:

                def _postprocess(res: tf_ext.EagerTensor) -> TFRunnableOutputType:
                    return t.cast("ext.NpNDArray", res.numpy())

        else:
            # if there are no output signature or more than one output
            # signatures, the post process function need to do casting
            # depends on the real output value each time

            def _postprocess(res: TFModelOutputType) -> TFRunnableOutputType:
                if isinstance(res, (tuple, list)):
                    return tuple(t.cast("ext.NpNDArray", r.numpy()) for r in res)
                else:
                    return t.cast("ext.NpNDArray", res.numpy())

        def _run_method(
            _runnable_self: TensorflowRunnable,
            *args: TFArgType,
            **kwargs: TFArgType,
        ) -> TFRunnableOutputType:
            if method_partial_kwargs is not None:
                kwargs = dict(method_partial_kwargs, **kwargs)

            try:
                res = raw_method(*args, **kwargs)
            except ValueError:
                # Tensorflow performs type checking implicitly if users decorate with `tf.function
                # or provide `tf_signatures` when calling `save_model()`. Type checking and
                # casting is deferred to after the `ValueError` is raised to optimize performance.
                sigs = get_input_signatures_v2(raw_method)
                if not sigs:
                    raise

                try:
                    casted_args = cast_py_args_to_tf_function_args(
                        sigs[0], *args, **kwargs
                    )
                except ValueError:
                    raise

                res = raw_method(*casted_args)
            return _postprocess(res)

        return _run_method

    def add_run_method(method_name: str, options: ModelSignature):
        def run_method(
            runnable_self: TensorflowRunnable, *args: TFArgType, **kwargs: TFArgType
        ) -> TFRunnableOutputType:
            _run_method = runnable_self.methods_cache.get(
                method_name
            )  # is methods_cache nessesary?
            if not _run_method:
                _run_method = _gen_run_method(runnable_self, method_name)
                runnable_self.methods_cache[method_name] = _run_method

            return _run_method(runnable_self, *args, **kwargs)

        TensorflowRunnable.add_method(
            run_method,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_run_method(method_name, options)

    return TensorflowRunnable


class TensorflowTensorContainer(
    DataContainer["tf_ext.EagerTensor", "tf_ext.EagerTensor"]
):
    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[tf_ext.EagerTensor], batch_dim: int = 0
    ) -> t.Tuple[tf_ext.EagerTensor, list[int]]:
        batch: tf_ext.EagerTensor = tf.concat(batches, axis=batch_dim)
        # TODO: fix typing mismatch @larme
        indices: list[int] = list(
            itertools.accumulate(subbatch.shape[batch_dim] for subbatch in batches)
        )  # type: ignore
        indices = [0] + indices
        return batch, indices

    @classmethod
    def batch_to_batches(
        cls, batch: tf_ext.EagerTensor, indices: t.Sequence[int], batch_dim: int = 0
    ) -> t.List[tf_ext.EagerTensor]:
        size_splits = [indices[i + 1] - indices[i] for i in range(len(indices) - 1)]
        return tf.split(batch, size_splits, axis=batch_dim)  # type: ignore

    @classmethod
    def to_payload(
        cls,
        batch: tf_ext.EagerTensor,
        batch_dim: int = 0,
    ) -> Payload:
        return cls.create_payload(
            pickle.dumps(batch),
            batch_size=batch.shape[batch_dim],
        )

    @classmethod
    def from_payload(
        cls,
        payload: Payload,
    ) -> tf_ext.EagerTensor:
        return pickle.loads(payload.data)

    @classmethod
    def batch_to_payloads(
        cls,
        batch: tf_ext.EagerTensor,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> t.List[Payload]:
        batches = cls.batch_to_batches(batch, indices, batch_dim)

        payloads = [cls.to_payload(subbatch) for subbatch in batches]
        return payloads

    @classmethod
    def from_batch_payloads(
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int = 0,
    ) -> t.Tuple[tf_ext.EagerTensor, t.List[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)


# NOTE: we only register the TensorflowTensorContainer as if eager execution is enabled.
if tf.executing_eagerly():
    DataContainerRegistry.register_container(
        LazyType("tensorflow.python.framework.ops", "Tensor"),
        LazyType("tensorflow.python.framework.ops", "Tensor"),
        TensorflowTensorContainer,
    )



from ._internal.frameworks.flax import ModelOptions
from ._internal.frameworks.flax import get
from ._internal.frameworks.flax import get_runnable
from ._internal.frameworks.flax import load_model
from ._internal.frameworks.flax import save_model

__all__ = ["load_model", "save_model", "get", "get_runnable", "ModelOptions"]



from __future__ import annotations

import typing as t
from functools import lru_cache
from http import HTTPStatus
from typing import TYPE_CHECKING

from deepmerge.merger import Merger

from bentoml.exceptions import InternalServerError
from bentoml.exceptions import InvalidArgument
from bentoml.exceptions import NotFound

from ...types import LazyType
from ...utils import bentoml_cattr
from .specification import Components
from .specification import Contact
from .specification import Info
from .specification import MediaType
from .specification import OpenAPISpecification
from .specification import Operation
from .specification import PathItem
from .specification import Reference
from .specification import Response
from .specification import Tag
from .utils import REF_PREFIX
from .utils import exception_components_schema
from .utils import exception_schema

if TYPE_CHECKING:
    from .. import Service
    from ..inference_api import InferenceAPI

SUCCESS_DESCRIPTION = "Successful Response"

INFRA_DECRIPTION = {
    "/healthz": "Health check endpoint. Expecting an empty response with status code <code>200</code> when the service is in health state. The <code>/healthz</code> endpoint is <b>deprecated</b>. (since Kubernetes v1.16)",
    "/livez": "Health check endpoint for Kubernetes. Healthy endpoint responses with a <code>200</code> OK status.",
    "/readyz": "A <code>200</code> OK status from <code>/readyz</code> endpoint indicated the service is ready to accept traffic. From that point and onward, Kubernetes will use <code>/livez</code> endpoint to perform periodic health checks.",
    "/metrics": "Prometheus metrics endpoint. The <code>/metrics</code> responses with a <code>200</code>. The output can then be used by a Prometheus sidecar to scrape the metrics of the service.",
}

__all__ = ["generate_spec"]

INFRA_TAG = Tag(
    name="Infrastructure",
    description="Common infrastructure endpoints for observability.",
)
APP_TAG = Tag(
    name="Service APIs", description="BentoML Service API endpoints for inference."
)

merger = Merger(
    # merge dicts
    [(dict, "merge")],
    # override all other types
    ["override"],
    # override conflicting types
    ["override"],
)


def make_api_path(api: InferenceAPI) -> str:
    return api.route if api.route.startswith("/") else f"/{api.route}"


@lru_cache(maxsize=1)
def make_infra_endpoints() -> dict[str, PathItem]:
    return {
        endpoint: PathItem(
            get=Operation(
                responses={"200": Response(description=SUCCESS_DESCRIPTION)},
                tags=[INFRA_TAG.name],
                description=INFRA_DECRIPTION[endpoint],
            )
        )
        for endpoint in INFRA_DECRIPTION
    }


def generate_service_components(svc: Service) -> Components:
    components: dict[str, t.Any] = {}
    for api in svc.apis.values():
        api_components = {}
        input_components = api.input.openapi_components()
        if input_components:
            merger.merge(api_components, input_components)
        output_components = api.output.openapi_components()
        if output_components:
            merger.merge(api_components, output_components)

        merger.merge(components, api_components)

    # merge exception at last
    merger.merge(components, {"schemas": exception_components_schema()})

    return Components(**components)


def generate_spec(svc: Service, *, openapi_version: str = "3.0.2"):
    """Generate a OpenAPI specification for a service."""
    mounted_app_paths = {}

    for app, _, _ in svc.mount_apps:
        if LazyType["fastapi.FastAPI"]("fastapi.FastAPI").isinstance(app):
            from fastapi.openapi.utils import get_openapi

            openapi = get_openapi(
                title=app.title,
                version=app.version,
                routes=app.routes,
            )

            mounted_app_paths.update(
                {
                    k: bentoml_cattr.structure(v, PathItem)
                    for k, v in openapi["paths"].items()
                }
            )

    return OpenAPISpecification(
        openapi=openapi_version,
        tags=[APP_TAG, INFRA_TAG],
        components=generate_service_components(svc),
        info=Info(
            title=svc.name,
            description=svc.doc,
            version=svc.tag.version if svc.tag and svc.tag.version else "None",
            contact=Contact(name="BentoML Team", email="contact@bentoml.com"),
        ),
        servers=[{"url": "."}],
        paths={
            # setup infra endpoints
            **make_infra_endpoints(),
            # setup inference endpoints
            **{
                make_api_path(api): PathItem(
                    post={
                        "responses": {
                            HTTPStatus.OK.value: api.output.openapi_responses(),
                            **{
                                ex.error_code.value: Response(
                                    description=filled.description,
                                    content={
                                        "application/json": MediaType(
                                            schema=Reference(
                                                f"{REF_PREFIX}{filled.title}"
                                            )
                                        )
                                    },
                                )
                                for ex in [
                                    InvalidArgument,
                                    NotFound,
                                    InternalServerError,
                                ]
                                for filled in exception_schema(ex)
                            },
                        },
                        "tags": [APP_TAG.name],
                        "consumes": [api.input.mime_type],
                        "produces": [api.output.mime_type],
                        "x-bentoml-name": api.name,
                        "summary": str(api),
                        "description": api.doc or "",
                        "requestBody": api.input.openapi_request_body(),
                        "operationId": f"{svc.name}__{api.name}",
                    },
                )
                for api in svc.apis.values()
            },
            **mounted_app_paths,
        },
    )



from __future__ import annotations

import json
import logging
import typing as t

import attr
from deepmerge.merger import Merger

from ...exceptions import BentoMLException
from ..tag import Tag
from ..utils import bentoml_cattr
from ..utils import first_not_none
from ..utils import resolve_user_filepath
from .config import get_rest_api_client
from .schemas import CreateDeploymentSchema
from .schemas import DeploymentListSchema
from .schemas import DeploymentMode
from .schemas import DeploymentSchema
from .schemas import DeploymentTargetCanaryRule
from .schemas import DeploymentTargetConfig
from .schemas import DeploymentTargetHPAConf
from .schemas import DeploymentTargetRunnerConfig
from .schemas import DeploymentTargetType
from .schemas import FullDeploymentSchema
from .schemas import UpdateDeploymentSchema

logger = logging.getLogger(__name__)

config_merger = Merger(
    # merge dicts
    type_strategies=[(dict, "merge")],
    # override all other types
    fallback_strategies=["override"],
    # override conflicting types
    type_conflict_strategies=["override"],
)


@attr.define
class Resource:
    @classmethod
    def for_hpa_conf(cls, **kwargs: t.Any) -> DeploymentTargetHPAConf:
        return bentoml_cattr.structure(kwargs, DeploymentTargetHPAConf)

    @classmethod
    def for_runner(cls, **kwargs: t.Any) -> DeploymentTargetRunnerConfig:
        exclusive_api_server_key = {
            v for v in kwargs if v not in attr.fields_dict(DeploymentTargetRunnerConfig)
        }
        return bentoml_cattr.structure(
            {k: v for k, v in kwargs.items() if k not in exclusive_api_server_key},
            DeploymentTargetRunnerConfig,
        )

    @classmethod
    def for_api_server(cls, **kwargs: t.Any) -> DeploymentTargetConfig:
        return bentoml_cattr.structure(kwargs, DeploymentTargetConfig)


class Deployment:
    @classmethod
    def _get_default_kube_namespace(
        cls,
        cluster_name: str,
        context: str | None = None,
    ) -> str:
        cloud_rest_client = get_rest_api_client(context)
        res = cloud_rest_client.get_cluster(cluster_name)
        if not res:
            raise BentoMLException("Cannot get default kube namespace")
        return res.config.default_deployment_kube_namespace

    @classmethod
    def _get_default_cluster(cls, context: str | None = None) -> str:
        cloud_rest_client = get_rest_api_client(context)
        res = cloud_rest_client.get_cluster_list(params={"count": 1})
        if not res:
            raise BentoMLException("Failed to get list of clusters.")
        if not res.items:
            raise BentoMLException("Cannot get default clusters.")
        return res.items[0].name

    @classmethod
    def _create_deployment(
        cls,
        create_deployment_schema: CreateDeploymentSchema,
        context: str | None = None,
        cluster_name: str | None = None,
    ) -> DeploymentSchema:
        cloud_rest_client = get_rest_api_client(context)
        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if create_deployment_schema.kube_namespace is None:
            create_deployment_schema.kube_namespace = cls._get_default_kube_namespace(
                cluster_name, context
            )
        for target in create_deployment_schema.targets:
            if (
                cloud_rest_client.get_bento(target.bento_repository, target.bento)
                is None
            ):
                raise BentoMLException(
                    f"Create deployment: {target.bento_repository}:{target.bento} does not exist"
                )
        if (
            cloud_rest_client.get_deployment(
                cluster_name,
                create_deployment_schema.kube_namespace,
                create_deployment_schema.name,
            )
            is not None
        ):
            raise BentoMLException("Create deployment: Deployment already exists")
        res = cloud_rest_client.create_deployment(
            cluster_name, create_deployment_schema
        )
        if res is None:
            raise BentoMLException("Create deployment request failed")
        logger.debug("Deployment Schema: %s", create_deployment_schema)
        return res

    @classmethod
    def _update_deployment(
        cls,
        deployment_name: str,
        update_deployment_schema: UpdateDeploymentSchema,
        kube_namespace: str | None = None,
        context: str | None = None,
        cluster_name: str | None = None,
    ) -> DeploymentSchema:
        cloud_rest_client = get_rest_api_client(context)
        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if kube_namespace is None:
            kube_namespace = cls._get_default_kube_namespace(cluster_name, context)
        for target in update_deployment_schema.targets:
            if (
                cloud_rest_client.get_bento(target.bento_repository, target.bento)
                is None
            ):
                raise BentoMLException(
                    f"Update deployment: {target.bento_repository}:{target.bento} does not exist"
                )
            cloud_rest_client.get_deployment(
                cluster_name,
                kube_namespace,
                deployment_name,
            )

        res = cloud_rest_client.update_deployment(
            cluster_name, kube_namespace, deployment_name, update_deployment_schema
        )
        if res is None:
            raise BentoMLException("Update deployment request failed")
        logger.debug("%s is created.", deployment_name)
        logger.debug("Deployment Schema: %s", update_deployment_schema)
        return res

    @classmethod
    def update(
        cls,
        deployment_name: str,
        bento: Tag | str | None = None,
        description: str | None = None,
        expose_endpoint: bool | None = None,
        cluster_name: str | None = None,
        kube_namespace: str | None = None,
        resource_instance: str | None = None,
        hpa_conf: DeploymentTargetHPAConf | None = None,
        runners_config: dict[str, DeploymentTargetRunnerConfig] | None = None,
        api_server_config: DeploymentTargetConfig | None = None,
        mode: DeploymentMode | None = None,
        type: DeploymentTargetType | None = None,
        context: str | None = None,
        labels: dict[str, str] | None = None,
        canary_rules: t.List[DeploymentTargetCanaryRule] | None = None,
        latest_bento: bool = False,
    ) -> DeploymentSchema:
        from bentoml import get as get_bento

        if mode is None:
            mode = DeploymentMode.Function
        if type is None:
            type = DeploymentTargetType.STABLE

        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if kube_namespace is None:
            kube_namespace = cls._get_default_kube_namespace(cluster_name, context)

        base_schema = cls.get(deployment_name, context, cluster_name, kube_namespace)
        # Deployment target always has length of 1
        if base_schema.latest_revision is None:
            raise BentoMLException(
                f"Deployment {deployment_name} has no latest revision"
            )
        deployment_target = base_schema.latest_revision.targets[0]

        if bento is None:
            # NOTE: bento.repository.name is the bento.name, and bento.name is the bento.version
            # from bentocloud to bentoml.Tag concept
            bento = deployment_target.bento.repository.name
        bento = Tag.from_taglike(bento)
        if latest_bento and bento.version is None or bento.version == "latest":
            bento = get_bento(bento).tag
        elif bento.version is None:
            bento.version = deployment_target.bento.name

        updated_config = bentoml_cattr.unstructure(deployment_target.config)
        if hpa_conf is not None:
            hpa_conf_dct = bentoml_cattr.unstructure(hpa_conf)
            if "hpa_conf" in updated_config:
                if updated_config["hpa_conf"] is None:
                    updated_config["hpa_conf"] = {}
                config_merger.merge(updated_config["hpa_conf"], hpa_conf_dct)
            if "runners" in updated_config and updated_config["runners"] is not None:
                for _, runner in updated_config["runners"].items():
                    if runner["hpa_conf"] is None:
                        runner["hpa_conf"] = {}
                    config_merger.merge(runner["hpa_conf"], hpa_conf_dct)
        if resource_instance is not None:
            updated_config["resource_instance"] = resource_instance
            if updated_config.get("runners") is not None:
                for runner in updated_config["runners"].values():
                    runner["resource_instance"] = resource_instance
        if expose_endpoint is not None:
            updated_config["enable_ingress"] = expose_endpoint

        if api_server_config is not None:
            if runners_config is not None:
                api_server_config.runners = runners_config
            config_merger.merge(
                updated_config, bentoml_cattr.unstructure(api_server_config)
            )
        elif runners_config is not None:
            config_merger.merge(
                updated_config,
                {
                    "runners": {
                        k: bentoml_cattr.unstructure(v)
                        for k, v in runners_config.items()
                    }
                },
            )

        dct_update: dict[str, t.Any] = {
            "mode": first_not_none(mode, base_schema.mode),
            "labels": first_not_none(
                [{"key": key, "value": value} for key, value in labels.items()]
                if labels
                else None,
                [bentoml_cattr.unstructure(i) for i in base_schema.labels],
            ),
            "description": description,
        }
        update_target = {
            "type": first_not_none(type, deployment_target.type),
            "bento": first_not_none(bento.version, deployment_target.bento.name),
            "bento_repository": first_not_none(
                bento.name, deployment_target.bento.repository.name
            ),
            "config": updated_config,
        }

        rules = first_not_none(
            [bentoml_cattr.unstructure(i) for i in canary_rules]
            if canary_rules
            else None,
            [bentoml_cattr.unstructure(i) for i in deployment_target.canary_rules]
            if deployment_target.canary_rules
            else None,
        )
        if rules:
            update_target["canary_rules"] = rules

        # update the target
        dct_update["targets"] = [update_target]

        return cls._update_deployment(
            deployment_name=deployment_name,
            update_deployment_schema=bentoml_cattr.structure(
                dct_update, UpdateDeploymentSchema
            ),
            context=context,
            cluster_name=cluster_name,
            kube_namespace=kube_namespace,
        )

    @classmethod
    def create(
        cls,
        deployment_name: str,
        bento: Tag | str,
        description: str | None = None,
        expose_endpoint: bool | None = None,
        cluster_name: str | None = None,
        kube_namespace: str | None = None,
        resource_instance: str | None = None,
        hpa_conf: DeploymentTargetHPAConf | None = None,
        runners_config: dict[str, DeploymentTargetRunnerConfig] | None = None,
        api_server_config: DeploymentTargetConfig | None = None,
        mode: DeploymentMode | None = None,
        type: DeploymentTargetType | None = None,
        context: str | None = None,
        labels: dict[str, str] | None = None,
        canary_rules: t.List[DeploymentTargetCanaryRule] | None = None,
    ) -> DeploymentSchema:
        if mode is None:
            mode = DeploymentMode.Function
        if type is None:
            type = DeploymentTargetType.STABLE
        bento_tag = Tag.from_taglike(bento)

        dct: dict[str, t.Any] = {
            "name": deployment_name,
            "kube_namespace": kube_namespace,
            "mode": mode,
            "description": description,
        }
        if labels:
            dct["labels"] = [
                {"key": key, "value": value} for key, value in labels.items()
            ]
        if api_server_config is None:
            _config: dict[str, t.Any] = {
                "runners": {
                    k: bentoml_cattr.unstructure(v) for k, v in runners_config.items()
                }
                if runners_config
                else None,
            }
        else:
            api_server_config.runners = runners_config
            _config = bentoml_cattr.unstructure(api_server_config)

        create_target: dict[str, t.Any] = {
            "type": type,
            "bento_repository": bento_tag.name,
            "bento": bento_tag.version,
            "config": _config,
        }
        if canary_rules:
            create_target["canary_rules"] = [
                bentoml_cattr.unstructure(i) for i in canary_rules
            ]

        # Only change the value by the top-level param if it is not provided already in api_server_config or runner_config
        if hpa_conf:
            hpa_conf_dct = bentoml_cattr.unstructure(hpa_conf)
            _config_hpa_conf = _config.get("hpa_conf", None)
            if _config_hpa_conf is None:
                _config_hpa_conf = {}
            for k, v in hpa_conf_dct.items():
                if k not in _config_hpa_conf:
                    _config_hpa_conf[k] = v
                else:
                    logger.warning(
                        "Key %s is already set in API server config and will not be overwritten with hpa_conf.%s",
                        k,
                        k,
                    )
            _config["hpa_conf"] = _config_hpa_conf
            if "runners" in _config and _config["runners"] is not None:
                _runner_config = _config["runners"]
                for runner in _runner_config.values():
                    _runner_hpa_conf = runner.get("hpa_conf", None)
                    if _runner_hpa_conf is None:
                        _runner_hpa_conf = {}
                    for k, v in hpa_conf_dct.items():
                        if k not in _runner_hpa_conf:
                            _runner_hpa_conf[k] = v
                        else:
                            logger.warning(
                                "Key %s is already set in runner config and will not be overwritten with hpa_conf.%s",
                                k,
                                k,
                            )
                    runner["hpa_conf"] = _runner_hpa_conf

        if resource_instance:
            if "resource_instance" not in _config:
                _config["resource_instance"] = resource_instance
            if "runners" in _config and _config["runners"] is not None:
                for runner in _config["runners"].values():
                    if "resource_instance" not in runner:
                        runner["resource_instance"] = resource_instance

        if expose_endpoint is not None and _config.get("enable_ingress", None) is None:
            _config["enable_ingress"] = expose_endpoint

        # setup the create target itself
        dct["targets"] = [create_target]

        return cls._create_deployment(
            context=context,
            cluster_name=cluster_name,
            create_deployment_schema=bentoml_cattr.structure(
                dct, CreateDeploymentSchema
            ),
        )

    @classmethod
    def list(
        cls,
        context: str | None = None,
        cluster_name: str | None = None,
        query: str | None = None,
        search: str | None = None,
        count: int | None = None,
        start: int | None = None,
    ) -> DeploymentListSchema:
        cloud_rest_client = get_rest_api_client(context)
        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if query or start or count or search:
            params = {"start": start, "count": count, "search": search, "q": query}
            res = cloud_rest_client.get_deployment_list(cluster_name, **params)
            if res is None:
                raise BentoMLException("List deployments request failed")
            return res
        else:
            all_deployment = cloud_rest_client.get_deployment_list(cluster_name)
            if all_deployment is None:
                raise BentoMLException("List deployments request failed")
            return all_deployment

    @classmethod
    def create_from_file(
        cls,
        path_or_stream: str | t.TextIO,
        path_context: str | None = None,
        context: str | None = None,
    ) -> DeploymentSchema:
        if isinstance(path_or_stream, str):
            real_path = resolve_user_filepath(path_or_stream, path_context)
            try:
                with open(real_path, "r") as file:
                    data = json.load(file)
            except FileNotFoundError:
                raise ValueError(f"File not found: {real_path}")
            except json.JSONDecodeError as e:
                raise ValueError(f"Error decoding JSON file: {real_path}\n{e}")
            except Exception as e:
                raise ValueError(
                    f"An error occurred while reading the file: {real_path}\n{e}"
                )
        else:
            # load the data from trusted stream
            data = json.load(path_or_stream)
        deployment_schema = bentoml_cattr.structure(data, FullDeploymentSchema)
        return cls._create_deployment(
            create_deployment_schema=deployment_schema,
            context=context,
            cluster_name=deployment_schema.cluster_name,
        )

    @classmethod
    def update_from_file(
        cls,
        path_or_stream: str | t.TextIO,
        path_context: str | None = None,
        context: str | None = None,
    ) -> DeploymentSchema:
        if isinstance(path_or_stream, str):
            real_path = resolve_user_filepath(path_or_stream, path_context)
            try:
                with open(real_path, "r") as file:
                    data = json.load(file)
            except FileNotFoundError:
                raise ValueError(f"File not found: {real_path}")
            except json.JSONDecodeError as e:
                raise ValueError(f"Error decoding JSON file: {real_path}\n{e}")
            except Exception as e:
                raise ValueError(
                    f"An error occurred while reading the file: {real_path}\n{e}"
                )
        else:
            data = json.load(path_or_stream)
        deployment_schema = bentoml_cattr.structure(data, FullDeploymentSchema)
        return cls._update_deployment(
            deployment_name=deployment_schema.name,
            update_deployment_schema=deployment_schema,
            context=context,
            cluster_name=deployment_schema.cluster_name,
            kube_namespace=deployment_schema.kube_namespace,
        )

    @classmethod
    def get(
        cls,
        deployment_name: str,
        context: str | None = None,
        cluster_name: str | None = None,
        kube_namespace: str | None = None,
    ) -> DeploymentSchema:
        cloud_rest_client = get_rest_api_client(context)
        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if kube_namespace is None:
            kube_namespace = cls._get_default_kube_namespace(cluster_name, context)
        res = cloud_rest_client.get_deployment(
            cluster_name, kube_namespace, deployment_name
        )
        if res is None:
            raise BentoMLException("Get deployment request failed")
        return res

    @classmethod
    def delete(
        cls,
        deployment_name: str,
        context: str | None = None,
        cluster_name: str | None = None,
        kube_namespace: str | None = None,
    ) -> DeploymentSchema:
        cloud_rest_client = get_rest_api_client(context)
        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if kube_namespace is None:
            kube_namespace = cls._get_default_kube_namespace(cluster_name, context)
        res = cloud_rest_client.get_deployment(
            cluster_name,
            kube_namespace,
            deployment_name,
        )
        if res is None:
            raise BentoMLException("Delete deployment: Deployment does not exist")

        res = cloud_rest_client.delete_deployment(
            cluster_name, kube_namespace, deployment_name
        )
        if res is None:
            raise BentoMLException("Delete deployment request failed")
        return res

    @classmethod
    def terminate(
        cls,
        deployment_name: str,
        context: str | None = None,
        cluster_name: str | None = None,
        kube_namespace: str | None = None,
    ) -> DeploymentSchema:
        cloud_rest_client = get_rest_api_client(context)
        if cluster_name is None:
            cluster_name = cls._get_default_cluster(context)
        if kube_namespace is None:
            kube_namespace = cls._get_default_kube_namespace(cluster_name, context)
        res = cloud_rest_client.get_deployment(
            cluster_name,
            kube_namespace,
            deployment_name,
        )
        if res is None:
            raise BentoMLException("Teminate deployment: Deployment does not exist")
        res = cloud_rest_client.terminate_deployment(
            cluster_name, kube_namespace, deployment_name
        )
        if res is None:
            raise BentoMLException("Terminate deployment request failed")
        return res
